"""Processing Task that creates a 2D map of sea ice variables."""

import datetime

import numpy as np
import iris
import cftime

from scriptengine.tasks.base.timing import timed_runner
import helpers.file_handling as helpers
from .map import Map

meta_dict = {
    'sivolu': 'Sea-Ice Volume per Area',
    'siconc': 'Sea-Ice Area Fraction',
}

class Si3HemisPointMonthMeanAllMeanMap(Map):
    """Si3HemisPointMonthMeanAllMeanMap Processing Task"""

    def __init__(self, parameters):
        super().__init__(
            parameters,
            required_parameters=['src', 'dst', 'hemisphere', 'varname']
            )

    @timed_runner
    def run(self, context):
        src = self.getarg('src', context)
        dst = self.getarg('dst', context)
        hemisphere = self.getarg('hemisphere', context)
        varname = self.getarg('varname', context)
        self.log_info(f"Create {varname} map for {hemisphere}ern hemisphere at {dst}.")
        self.log_debug(f"Source file(s): {src}")

        if varname not in meta_dict:
            self.log_warning((
                f"'varname' must be one of the following: {meta_dict.keys()} "
                f"Diagnostic will not be treated, returning now."
            ))
            return
        if not hemisphere in ('north', 'south'):
            self.log_warning((
                f"'hemisphere' must be 'north' or 'south' but is '{hemisphere}'."
                f"Diagnostic will not be treated, returning now."
            ))
            return
        if not self.correct_file_extension(dst):
            return

        month_cube = helpers.load_input_cube(src, varname)
        # Remove auxiliary time coordinate
        month_cube.remove_coord(month_cube.coord('time', dim_coords=False))
        month_cube = month_cube[0]
        time_coord = month_cube.coord('time')
        time_coord.bounds = self.get_time_bounds(time_coord)
        latitudes = np.broadcast_to(month_cube.coord('latitude').points, month_cube.shape)
        if hemisphere == "north":
            month_cube.data = np.ma.masked_where(latitudes < 0, month_cube.data)
        else:
            month_cube.data = np.ma.masked_where(latitudes > 0, month_cube.data)

        month_cube.long_name = f"{meta_dict[varname]} {hemisphere} {self.get_month(time_coord)}"
        month_cube.data = np.ma.masked_equal(month_cube.data, 0)

        month_cube.data = month_cube.data.astype('float64')
        comment = f"Simulation Average of {meta_dict[varname]} / **{varname}** on {hemisphere}ern hemisphere."
        month_cube = helpers.set_metadata(
            month_cube,
            title=f'{month_cube.long_name} (Climatology)',
            comment=comment,
            map_type='polar ice sheet',
        )
        time_coord.climatological = True
        month_cube = self.set_cell_methods(month_cube, hemisphere)

        self.save(month_cube, dst)

    def get_time_bounds(self, time_coord):
        """
        Get contiguous time bounds for sea ice maps

        Creates new time bounds [
            [01-01-current year, 01-01-next year],
        ]
        """
        dt_object = cftime.num2pydate(time_coord.points[0], time_coord.units.name)
        start = datetime.datetime(dt_object.year, 1, 1)
        end = datetime.datetime(start.year + 1, 1, 1)
        start_seconds = cftime.date2num(start, time_coord.units.name)
        end_seconds = cftime.date2num(end, time_coord.units.name)
        new_bounds = np.array([[start_seconds, end_seconds]])
        return new_bounds

    def get_month(self, time_coord):
        """
        Returns the month of [0] in time_coord.points as a string
        """
        dt_object = cftime.num2pydate(time_coord.points[0], time_coord.units.name)
        return dt_object.strftime('%B')
    
    def set_cell_methods(self, cube, hemisphere):
        """Set the correct cell methods."""
        self.log_debug("Setting cell methods.")
        cube.cell_methods = ()
        cube.add_cell_method(iris.coords.CellMethod('mean over years', coords='time'))
        cube.add_cell_method(iris.coords.CellMethod(
            'point', coords='latitude', intervals=f'{hemisphere}ern hemisphere'
            ))
        cube.add_cell_method(iris.coords.CellMethod('point', coords='longitude'))
        return cube



"""Presentation Task that uploads Data and Plots to the EC-Earth dev portal."""

import os
import urllib.parse
import requests

import redminelib
import redminelib.exceptions

import jinja2
import yaml
import iris

from scriptengine.tasks.base import Task
from scriptengine.tasks.base.timing import timed_runner
from scriptengine.jinja import filters as j2filters
from helpers.presentation_objects import PresentationObject
from helpers.exceptions import PresentationException
from helpers.file_handling import ChangeDirectory

class Redmine(Task):
    """Redmine Presentation Task"""
    def __init__(self, parameters):
        required = [
            "src",
            "local_dst",
            "subject",
            "template",
            "api_key",
        ]
        super().__init__(__name__, parameters, required_parameters=required)

    @timed_runner
    def run(self, context):
        sources = self.getarg('src', context)
        dst_folder = self.getarg('local_dst', context)
        issue_subject = self.getarg('subject', context, parse_yaml=False)
        template_path = self.getarg('template', context)
        key = self.getarg('api_key', context)
        self.log_info(f"Create Redmine issue '{issue_subject}'.")
        self.log_debug(f"Template: {template_path}, Source File(s): {sources}")

        presentation_list = self.get_presentation_list(sources, dst_folder)
        redmine_template = self.get_template(context, template_path)
        redmine_template.globals['urlencode'] = urllib.parse.quote
        issue_description = redmine_template.render(presentation_list=presentation_list)
        with ChangeDirectory(dst_folder):
            with open("./issue_description.txt", 'w') as outfile:
                outfile.write(issue_description)

        url = 'https://dev.ec-earth.org'
        redmine = redminelib.Redmine(url, key=key)

        self.log_debug("Connecting to Redmine.")
        issue = self.get_issue(redmine, issue_subject)
        if issue is None:
            return

        self.log_debug("Updating the issue description.")
        issue.description = ""
        for line in issue_description:
            issue.description += line

        self.log_debug("Uploading attachments.")
        issue.uploads = []
        for item in presentation_list:
            if item['presentation_type'] == 'image':
                file_name = os.path.basename(item['path'])
                try:
                    for attachment in issue.attachments or []:
                        if attachment.filename == file_name:
                            redmine.attachment.delete(attachment.id)
                except redminelib.exceptions.ResourceNotFoundError:
                    pass
                issue.uploads.append({'filename': file_name, 'path': f"{dst_folder}/{file_name}"})
        self.log_debug("Saving issue.")
        issue.save()

    def get_presentation_list(self, sources, dst_folder):
        """create a list of presentation objects"""
        self.log_debug("Getting list of presentation objects.")
        presentation_list = []
        for src in sources:
            try:
                try:
                    pres_object = PresentationObject(dst_folder, **src)
                except TypeError:
                    pres_object = PresentationObject(dst_folder, src)
                self.log_debug(f"Loading {pres_object.loader.diag_type} diagnostic from {pres_object.loader.path}.")
                presentation_list.append(pres_object.create_dict())
            except PresentationException as msg:
                self.log_warning(f"Can not present diagnostic: {msg}")
        return presentation_list

    def get_template(self, context, template):
        """get Jinja2 template file"""
        search_path = ['.', 'templates']
        if "_se_cmd_cwd" in context:
            search_path.extend([context["_se_cmd_cwd"],
                                os.path.join(context["_se_cmd_cwd"], "templates")])
        self.log_debug(f"Search path for template: {search_path}")

        loader = jinja2.FileSystemLoader(search_path)
        environment = jinja2.Environment(loader=loader)
        for name, function in j2filters().items():
            environment.filters[name] = function
        return environment.get_template(template)

    def get_issue(self, redmine, issue_subject):
        """Connect to Redmine server, find and return issue corresponding to the experiment ID"""

        project_identifier = 'ec-earth-experiments'
        tracker_identifier = 15 # 'Experiment'
        status_identifier = 14 # 'Ongoing'
        priority_identifier = 2 # 'Medium'

        # Find issue or create if none exists
        filtered_issues = redmine.issue.filter(project_id=project_identifier, tracker_id=tracker_identifier)
        try:
            for issue in filtered_issues:
                if issue.subject == issue_subject:
                    break
            else:
                issue = redmine.issue.new()
                issue.subject = issue_subject
                issue.project_id = project_identifier
                issue.tracker_id = tracker_identifier
                issue.status_id = status_identifier
                issue.priority_id = priority_identifier
                issue.assigned_to_id = redmine.auth().id
                issue.is_private = False
            return issue
        except redminelib.exceptions.AuthError:
            self.log_warning('Could not log in to Redmine server (AuthError)')
            return
        except requests.exceptions.ConnectionError:
            self.log_warning('Could not log in to Redmine server (ConnectionError)')
            return



"""Tests for scriptengine/tasks/monitoring/markdown.py"""

from unittest.mock import patch, Mock
import os

import pytest
import iris
import yaml

from scriptengine.tasks.ecearth.monitoring.markdown import Markdown

def test_markdown_output_full(tmpdir):
    init = {
        "src": [str(tmpdir) + "/test.yml"],
        "dst": str(tmpdir),
        "template": './docs/templates/markdown_template.md.j2',
    }
    with open(init['src'][0], 'w') as file:
        yaml.dump(init, file)
    markdown_output = Markdown(init)

    markdown_output.run(init)
    assert os.path.isfile(init['dst'] + "/summary.md")

def test_markdown_presentation_list(tmpdir):
    init = {
        "src": [str(tmpdir) + "/test.yml"],
        "dst": str(tmpdir),
        "template": './docs/templates/markdown_template.md.j2',
    }
    markdown = Markdown(init)
    test_sources = [
        'path.txt',
        {'path': 'path.txt'},
        'path.yml'
    ]
    with patch.object(markdown, 'log_warning') as mock:
        result = markdown.get_presentation_list(test_sources, str(tmpdir))
    mock.assert_called_with("Can not present diagnostic: File not found! Ignoring path.yml")
    assert result == []



"""
Initialize presentation objects for visualization
"""

import os
import math
import abc

import yaml
import cftime
import matplotlib.pyplot as plt
import imageio
import iris
import iris.quickplot as qplt

from helpers.file_handling import ChangeDirectory
import helpers.exceptions as exceptions
import helpers.map_type_handling as type_handling

class PresentationObject:
    def __init__(self, dst_folder, path, **kwargs):
        self.dst_folder = dst_folder
        self.path = path
        self.custom_input = kwargs
        self.loader = get_loader(path)

    def create_dict(self):
        loaded_dict = self.loader.load(self.dst_folder, **self.custom_input)
        return {'presentation_type': self.loader.pres_type, **loaded_dict}

def get_loader(path):
    if path.endswith('.yml') or path.endswith('.yaml'):
        return ScalarLoader(path)
    if path.endswith('.nc'):
        try:
            loaded_cube = iris.load_cube(path)
        except OSError:
            raise exceptions.PresentationException(f"File not found! Ignoring {path}")
        if loaded_cube.attributes["diagnostic_type"] == "time series":
            return TimeseriesLoader(path, loaded_cube)
        if loaded_cube.attributes["diagnostic_type"] == "map":
            return MapLoader(path, loaded_cube)
        if loaded_cube.attributes["diagnostic_type"] == "temporal map":
            return TemporalmapLoader(path, loaded_cube)
        raise exceptions.PresentationException(f"Invalid diagnostic type {loaded_cube.attributes['diagnostic_type']}")
    raise exceptions.PresentationException(f"Invalid file extension of {path}")

class PresentationObjectLoader:
    def __init__(self, path):
        self.path = path
        self.diag_type = 'invalid'
        self.pres_type = 'invalid'

    def load(self, **kwargs):
        raise NotImplementedError

class ScalarLoader(PresentationObjectLoader):
    def __init__(self, path):
        self.path = path
        self.diag_type = 'scalar'
        self.pres_type = 'text'

    def load(self, *args, **kwargs):
        try:
            with open(self.path) as yml_file:
                loaded_dict = yaml.load(yml_file, Loader=yaml.FullLoader)
            return loaded_dict
        except FileNotFoundError:
            raise exceptions.PresentationException(f"File not found! Ignoring {self.path}")


class TimeseriesLoader(PresentationObjectLoader):
    def __init__(self, path, cube):
        self.path = path
        self.cube = cube
        self.diag_type = 'time series'
        self.pres_type = 'image'

    def load(self, dst_folder, **kwargs):
        """
        Load time series diagnostic and call plot creator.
        """
        # get file name without extension
        base_name = os.path.splitext(os.path.basename(self.path))[0]
        dst_file = f"./{base_name}.png"

        x_coord = self.cube.coords()[0]
        if "second since" in x_coord.units.name or "hour since" in x_coord.units.name:
            dates = cftime.num2pydate(x_coord.points, x_coord.units.name)
            coord_points = format_dates(dates)
        else:
            coord_points = x_coord.points
        
        fig = plt.figure(figsize=(6, 4), dpi=150)
        ax = fig.add_subplot(1, 1, 1)
        ax.plot(coord_points, self.cube.data, marker='o')
        if "second since" in x_coord.units.name:
            fig.autofmt_xdate()
        minor_step = math.ceil(len(coord_points) / 40)
        if len(coord_points) < 10:
            major_step = minor_step
        elif len(coord_points) < 20:
            major_step = 2*minor_step
        else:
            major_step = 3*minor_step
        ax.set_xticks(coord_points[::major_step])
        ax.set_xticks(coord_points[::minor_step], minor=True)
        ax.set_xticklabels(coord_points[::major_step])
        ax.ticklabel_format(axis='y', style='sci', scilimits=(-3, 6), useOffset=False, useMathText=True)
        ax.set_ylim(kwargs.get('value_range', [None, None]))
        ax.set_title(format_title(self.cube.long_name))
        ax.set_xlabel(format_title(x_coord.name()))
        ax.set_ylabel(format_title(self.cube.long_name, self.cube.units))
        plt.tight_layout()
        with ChangeDirectory(dst_folder):
            fig.savefig(dst_file, bbox_inches="tight")
            plt.close(fig)

        image_dict = {
            'title': self.cube.attributes['title'],
            'path': dst_file,
            'comment': self.cube.attributes['comment'],
        }
        return image_dict

class MapLoader(PresentationObjectLoader):
    def __init__(self, path, cube):
        self.path = path
        self.cube = cube
        self.diag_type = 'map'
        self.pres_type = 'image'

    def load(self, dst_folder, **kwargs):
        """
        Load map diagnostic and determine map type.
        """
        # get file name without extension
        base_name = os.path.splitext(os.path.basename(self.path))[0]
        map_type = self.cube.attributes['map_type']
        map_handler = type_handling.function_mapper(map_type)
        if map_handler is None:
            raise exceptions.InvalidMapTypeException(map_type)
        
        unit_text = f"{format_units(self.cube.units)}"
        time_coord = self.cube.coord('time')
        time_bounds = time_coord.bounds[0]
        dates = cftime.num2pydate(time_bounds, time_coord.units.name)
        plot_title = format_title(self.cube.long_name)
        date_title = f"{dates[0].strftime('%Y')} - {dates[-1].strftime('%Y')}"
        fig = map_handler(
            self.cube,
            title=plot_title,
            dates=date_title,
            units=unit_text,
            **kwargs,
        )
        dst_file = f"./{base_name}.png"
        with ChangeDirectory(dst_folder):
            fig.savefig(dst_file, bbox_inches="tight")
            plt.close(fig)

        image_dict = {
            'title': self.cube.attributes['title'],
            'path': dst_file,
            'comment': self.cube.attributes['comment'],
        }
        return image_dict

class TemporalmapLoader(PresentationObjectLoader):
    def __init__(self, path, cube):
        self.path = path
        self.cube = cube
        self.diag_type = 'temporal map'
        self.pres_type = 'image'

    def load(self, dst_folder,**kwargs):
        """
        Load map diagnostic and determine map type.
        """
        # get file name without extension
        base_name = os.path.splitext(os.path.basename(self.path))[0]
        map_type = self.cube.attributes['map_type']
        map_handler = type_handling.function_mapper(map_type)
        if map_handler is None:
            raise exceptions.InvalidMapTypeException(map_type)
        
        png_dir = f"{base_name}_frames"
        number_of_time_steps = len(self.cube.coord('time').points)
        with ChangeDirectory(dst_folder):
            if not os.path.isdir(png_dir):
                os.mkdir(png_dir)
            number_of_pngs = len(os.listdir(png_dir))

        unit_text = f"{format_units(self.cube.units)}"
        dst_file = f"./{base_name}.gif"
        with ChangeDirectory(f"{dst_folder}/{png_dir}"):
            for time_step in range(number_of_pngs, number_of_time_steps):
                time_coord = self.cube.coord('time')
                time_bounds = time_coord.bounds[time_step]
                dates = cftime.num2pydate(time_bounds, time_coord.units.name)
                date_title = f"{dates[0].strftime('%Y')}"
                plot_title = format_title(self.cube.long_name)
                fig = map_handler(
                    self.cube[time_step],
                    title=plot_title,
                    dates=date_title,
                    units=unit_text,
                    **kwargs,
                )
                fig.savefig(f"./{base_name}-{time_step:03}.png", bbox_inches="tight")
                plt.close(fig)   
            images = []
            for file_name in sorted(os.listdir(".")):
                images.append(imageio.imread(file_name))
            imageio.mimsave(f'.{dst_file}', images, fps=2)

        image_dict = {
            'title': self.cube.attributes['title'],
            'path': dst_file,
            'comment': self.cube.attributes['comment'],
        }
        return image_dict

def format_title(name, units=None):
    """
    Create Plot/Axis Title from Iris cube/coordinate

    Slight variance on _title() in iris/quickplot.py.
    """
    title = name.replace("_", " ").title()
    unit_text = format_units(units)
    if unit_text and unit_text != "1":
        title += " / {}".format(unit_text)
    return title

def format_units(units):
    """Format Cube Units as String"""
    if not (
            units is None
            or units.is_unknown()
            or units.is_no_unit()
        ):
        if qplt._use_symbol(units):
            return units.symbol
        else:
            return units
    else:
        return None

def format_dates(dates):
    fmt_dates = []
    for date in dates:
        fmt_dates.append(date.year)
    if len(set(fmt_dates)) != len(fmt_dates):
        fmt_dates = []
        for date in dates:
            fmt_dates.append(date.strftime("%Y-%m"))
    return fmt_dates



"""Tests for scriptengine/tasks/ecearth/monitoring/map.py"""

from scriptengine.tasks.ecearth.monitoring.map import Map

def test_map_dst_error():
    test_map = Map({})
    assert not test_map.correct_file_extension("test.yml")



"""Tests for scriptengine/tasks/ecearth/monitoring/nemo_all_mean_map.py"""

import iris

from scriptengine.tasks.ecearth.monitoring.nemo_all_mean_map import NemoAllMeanMap

def test_nemo_all_mean_map_once(tmpdir):
    init = {
        "src": [
            './tests/testdata/NEMO_output_sivolu-199003.nc',
            './tests/testdata/NEMO_output_sivolu-199009.nc',
            ],
        "dst": str(tmpdir) + '/test.nc',
        "varname": "sivolu",
    }
    ocean_map = NemoAllMeanMap(init)
    ocean_map.run(init)
    cube = iris.load_cube(init['dst'])
    assert cube.attributes['map_type'] == 'global ocean'
    assert cube.attributes['diagnostic_type'] == 'map'
    assert cube.coord('time').climatological
    assert len(cube.coord('time').points) == 1

def test_nemo_all_mean_map_twice(tmpdir):
    init_a = {
        "src": ['./tests/testdata/NEMO_output_sivolu-199003.nc'],
        "dst": str(tmpdir) + '/test.nc',
        "varname": "sivolu",
    }
    ocean_map = NemoAllMeanMap(init_a)
    ocean_map.run(init_a)
    init_b = {
        "src": ['./tests/testdata/NEMO_output_sivolu-199103.nc'],
        "dst": str(tmpdir) + '/test.nc',
        "varname": "sivolu",
    }
    ocean_map = NemoAllMeanMap(init_b)
    ocean_map.run(init_b)
    cube = iris.load_cube(init_b['dst'])
    assert cube.attributes['map_type'] == 'global ocean'
    assert cube.attributes['diagnostic_type'] == 'map'
    assert cube.coord('time').climatological
    assert len(cube.coord('time').points) == 1



"""ScriptEngine-Tasks-ECEarth exceptions
"""

class MonitoringException(Exception):
    """Exceptions from the Monitoring Framework."""

class PresentationException(MonitoringException):
    """Raise this exception when errors occur during a presentation task."""

class InvalidMapTypeException(PresentationException):
    """Raise this exception when map type does not exist."""


"""Tests for helpers/file_handling.py"""

import os

import pytest
import numpy as np
import iris

import helpers.file_handling as file_handling

def test_change_directory(tmpdir):
    cwd = os.getcwd()
    nwd = tmpdir.mkdir('temp')
    with file_handling.ChangeDirectory(nwd):
        assert os.getcwd() == nwd
    assert os.getcwd() == cwd

def test_compute_spatial_weights(tmpdir):
    e1t_cube = iris.cube.Cube([1], var_name='e1t')
    e2t_cube = iris.cube.Cube([2], var_name='e2t')
    cube_list = iris.cube.CubeList([e1t_cube, e2t_cube])
    domain_path = str(tmpdir + "/temp.nc")
    iris.save(cube_list, domain_path)
    result = np.array([[[2]], [[2]], [[2]]])
    shape = result.shape
    assert file_handling.compute_spatial_weights(domain_path, shape, 'T').all() == result.all()

def test_load_input_cube():
    src = "./tests/testdata/tos_nemo_all_mean_map.nc"
    varname = "tos"
    assert isinstance(file_handling.load_input_cube(src, varname), iris.cube.Cube)

def test_set_metadata():
    cube = iris.cube.Cube([1])
    cube.attributes = {
        'description': None,
        'interval_operation': None,
        'interval_write': None,
        'name': None,
        'online_operation': None,
    }
    updated_cube = file_handling.set_metadata(cube)
    assert updated_cube.attributes == {
        'source': 'EC-Earth 4',
        'Conventions': 'CF-1.8'
        }
    new_metadata = {
        'title': 'Title',
        'comment': 'Comment',
        'diagnostic_type': 'Type',
        'source': 'EC-Earth 4',
        'Conventions': 'CF-1.8',
        'custom': 'Custom',
    }
    updated_cube = file_handling.set_metadata(updated_cube, **new_metadata)
    assert updated_cube.attributes == new_metadata



"""Processing Task that calculates the seasonal cycle of sea ice variables in one leg."""

import warnings

import iris
import numpy as np
import cf_units

from scriptengine.tasks.base.timing import timed_runner
import helpers.file_handling as helpers
from .timeseries import Timeseries

meta_dict = {
    'sivolu':
        {
            'long_name': 'Sea-Ice Volume',
            'standard_name': 'sea_ice_volume',
            'var_name': 'sivol',
            'old_unit': 'm3',
            'new_unit': '1e3 km3',
        },
    'siconc':
        {
            'long_name': 'Sea-Ice Area',
            'standard_name': 'sea_ice_area',
            'var_name': 'siarea',
            'old_unit': 'm2',
            'new_unit': '1e6 km2',
        },
}

class Si3HemisSumMonthMeanTimeseries(Timeseries):
    """Si3HemisSumMonthMeanTimeseries Processing Task"""
    def __init__(self, parameters):
        super().__init__(
            {**parameters, 'title': None, 'coord_value': None, 'data_value': None},
            required_parameters=['src', 'domain', 'hemisphere', 'varname']
            )

    @timed_runner
    def run(self, context):
        dst = self.getarg('dst', context)
        varname = self.getarg('varname', context)
        hemisphere = self.getarg('hemisphere', context)

        self.log_info(f"Create {varname} time series for {hemisphere}ern hemisphere at {dst}.")

        if varname not in meta_dict:
            self.log_warning((
                f"'varname' must be one of the following: {meta_dict.keys()} "
                f"Diagnostic will not be treated, returning now."
            ))
            return
        long_name = meta_dict[varname]['long_name']

        src = self.getarg('src', context)
        domain = self.getarg('domain', context)
        
        self.log_debug(f"Domain: {domain}, Source file(s): {src}")

        if not hemisphere in ('north', 'south'):
            self.log_warning((
                f"'hemisphere' must be 'north' or 'south' but is '{hemisphere}'."
                f"Diagnostic will not be treated, returning now."
            ))
            return
        if not self.correct_file_extension(dst):
            return

        leg_cube = helpers.load_input_cube(src, varname)
        cell_weights = helpers.compute_spatial_weights(domain, leg_cube.shape, 'T')
        latitudes = np.broadcast_to(leg_cube.coord('latitude').points, leg_cube.shape)
        if hemisphere == "north":
            leg_cube.data = np.ma.masked_where(latitudes < 0, leg_cube.data)
        else:
            leg_cube.data = np.ma.masked_where(latitudes > 0, leg_cube.data)
        with warnings.catch_warnings():
            # Suppress warning about insufficient metadata.
            warnings.filterwarnings(
                'ignore',
                "Collapsing a multi-dimensional coordinate.",
                UserWarning,
                )
            hemispheric_sum = leg_cube.collapsed(
                ['latitude', 'longitude'],
                iris.analysis.SUM,
                weights=cell_weights,
                )

        # Remove auxiliary time coordinate
        hemispheric_sum.remove_coord(hemispheric_sum.coord('time', dim_coords=False))
        hemispheric_sum.standard_name = meta_dict[varname]['standard_name']
        hemispheric_sum.units = cf_units.Unit(meta_dict[varname]['old_unit'])
        hemispheric_sum.convert_units(meta_dict[varname]['new_unit'])
        hemispheric_sum.long_name = f"{long_name} {hemisphere.capitalize()}"
        hemispheric_sum.var_name = meta_dict[varname]['var_name'] + hemisphere[0]

        metadata = {
            'comment': (f"Sum of {long_name} / **{varname}** on {hemisphere}ern hemisphere."),
            'title': f"{long_name} (Seasonal Cycle)",
        }
        hemispheric_sum = helpers.set_metadata(hemispheric_sum, **metadata)
        hemispheric_sum = self.set_cell_methods(hemispheric_sum, hemisphere)
        self.save(hemispheric_sum, dst)
    
    def set_cell_methods(self, cube, hemisphere):
        """Set the correct cell methods."""
        cube.cell_methods = ()
        cube.add_cell_method(iris.coords.CellMethod('point', coords='time'))
        cube.add_cell_method(
            iris.coords.CellMethod('sum', coords='area', intervals=f'{hemisphere}ern hemisphere')
            )
        return cube



"""Module containing plot functions for different map types."""

import warnings

import iris
import iris.plot as iplt
import matplotlib.pyplot as plt
import cftime
import cf_units as unit
import cartopy.crs as ccrs

def function_mapper(map_type_string):
    mapper = {
        'global ocean': global_ocean_plot,
        'global atmosphere': global_atmosphere_plot,
        'polar ice sheet': polar_ice_sheet_plot,
    }
    return mapper.get(map_type_string, None)

def global_ocean_plot(cube, title=None, dates=None, units=None, value_range=[None, None], colormap='RdBu_r', **kwargs):
    """Map Type Handling for Global Ocean Maps"""
    fig = plt.figure(figsize=(6, 4), dpi=150)
    fig.suptitle(title)
    ax = fig.add_subplot(
        1, 1, 1,
        projection=ccrs.PlateCarree(),
        facecolor='#d3d3d3',
        )
    with warnings.catch_warnings():
        warnings.filterwarnings(
            action='ignore',
            message="Coordinate system of latitude and longitude ",
            category=UserWarning,
            )
        warnings.filterwarnings(
            action='ignore',
            message="Coordinate 'projection_._coordinate' is",
            category=UserWarning,
            )
        projected_cube, _ = iris.analysis.cartography.project(
            cube,
            ccrs.PlateCarree(),
            nx=800,
            ny=400,
            )
        im = iplt.pcolormesh(
            projected_cube,
            axes=ax,
            vmin=value_range[0],
            vmax=value_range[1],
            cmap=colormap,
            )
    cbar = fig.colorbar(im, orientation='horizontal')
    cbar.set_label(units)
    ax.set_title(dates, fontdict={'fontsize': 8, 'fontweight': 'medium'})
    ax.coastlines()
    return fig

def global_atmosphere_plot(cube, title=None, dates=None, units=None, value_range=[None, None], colormap='RdBu_r', **kwargs):
    """Map Type Handling for Global Atmosphere Maps"""
    fig = plt.figure(figsize=(6, 4), dpi=150)
    fig.suptitle(title)
    ax = fig.add_subplot(
        1, 1, 1,
        projection=ccrs.PlateCarree(),
        facecolor='#d3d3d3',
        )
    longitude = cube.coord('longitude').points
    latitude = cube.coord('latitude').points
    data = cube.data
    im = plt.scatter(
        longitude,
        latitude,
        s=1,
        c=data,
        axes=ax,
        vmin=value_range[0],
        vmax=value_range[1],
        cmap=colormap,
        transform=ccrs.PlateCarree(),
    )
    cbar = fig.colorbar(im, orientation='horizontal')
    cbar.set_label(units)
    ax.set_title(dates, fontdict={'fontsize': 8, 'fontweight': 'medium'})
    ax.coastlines()
    return fig

def polar_ice_sheet_plot(cube, title=None, dates=None, units=None, value_range=[None, None], colormap='RdBu_r', **kwargs):
    fig = plt.figure(figsize=(6,4), dpi=150)
    fig.suptitle(title)
    if "north" in cube.long_name or "North" in cube.long_name:
        center = 90.0
    else:
        center = -90.0
    ax = fig.add_subplot(
        1, 1, 1,
        projection=ccrs.Orthographic(central_latitude=center),
        facecolor='#d3d3d3',
        )
    with warnings.catch_warnings():
        warnings.filterwarnings(
            action='ignore', 
            message="Coordinate system of latitude and longitude ", 
            category=UserWarning,
            )
        warnings.filterwarnings(
            action='ignore', 
            message="Coordinate 'projection_._coordinate' is", 
            category=UserWarning,
            )
        projected_cube, _ = iris.analysis.cartography.project(
            cube, 
            ccrs.PlateCarree(),
            nx=800,
            ny=400,
            )
        im = iplt.pcolormesh(
            projected_cube,
            axes=ax,
            vmin=value_range[0],
            vmax=value_range[1],
            cmap=colormap,
            )
    bar = fig.colorbar(im, orientation='horizontal')
    if units:
        bar.set_label(units)
    ax.set_title(dates, fontdict={'fontsize': 8, 'fontweight': 'medium'})
    ax.coastlines()
    return fig



"""Tests for helpers/presentation_objects.py"""

import datetime

import pytest
import cf_units
import iris
import matplotlib.pyplot as plt
import yaml

import helpers.presentation_objects as po
import helpers.exceptions as exceptions

def test_format_title():
    assert po.format_title("test_title") == "Test Title"
    test_unit = cf_units.Unit("1")
    assert po.format_title("test_title", units=test_unit) == "Test Title"
    test_unit = cf_units.Unit("kilometers")
    assert po.format_title("test_title", units=test_unit) == "Test Title / 1000 m"

unit = [
    None,
    cf_units.Unit("1"),
    cf_units.Unit("kilometers"),
    cf_units.Unit("degC"),
    cf_units.Unit("no_unit")
]
expected_result = [
    None,
    "1",
    "1000 m",
    "degC",
    None
]

@pytest.mark.parametrize("unit, expected_result", zip(unit, expected_result))
def test_format_units(unit, expected_result):
    assert expected_result == po.format_units(unit)


def test_format_dates():
    current_date = datetime.datetime.now()
    changed_year = datetime.datetime(current_date.year - 1, current_date.month, 25)
    changed_month = datetime.datetime(current_date.year, (current_date.month + 1) % 12, 25)
    assert [current_date.year, changed_year.year] == po.format_dates([current_date, changed_year])
    assert [current_date.strftime("%Y-%m"), changed_month.strftime("%Y-%m")] == po.format_dates([current_date, changed_month])

def test_get_loader_extension():
    src = ["test.nc", "test.txt"]
    error_messages = [
        f"File not found! Ignoring {src[0]}",
        f"Invalid file extension of {src[1]}",
    ]
    
    for src, msg in zip(src, error_messages):
        with pytest.raises(exceptions.PresentationException, match=msg):
            po.get_loader(src)

def test_get_loader_invalid_diagnostic_type(tmpdir):
    src = str(tmpdir) + "/test.nc"
    cube = iris.cube.Cube([0], attributes={'diagnostic_type': 'invalid'})
    iris.save(cube, src)
    msg = f"Invalid diagnostic type {cube.attributes['diagnostic_type']}"
    with pytest.raises(exceptions.PresentationException, match=msg):
            po.get_loader(src)

def test_map_object(tmpdir, monkeypatch):
    path = './tests/testdata/tos_nemo_all_mean_map.nc'
    dst_folder = str(tmpdir)
    def mockreturn(*args, **kwargs):
        return plt.figure()
    
    map_object = po.PresentationObject(dst_folder, path)
    assert isinstance(map_object.loader, po.MapLoader)
    
    monkeypatch.setattr("helpers.map_type_handling.global_ocean_plot", mockreturn)
    result = map_object.create_dict()
    cube = iris.load_cube(path)
    expected_result = {
        'title': cube.attributes['title'],
        'path': './tos_nemo_all_mean_map.png',
        'comment': cube.attributes['comment'],
        'presentation_type': 'image',
    }
    assert result == expected_result

def test_temporalmap_object(tmpdir, monkeypatch):
    path = './tests/testdata/tos_nemo_year_mean_temporalmap.nc'
    dst_folder = str(tmpdir)
    def mockreturn(*args, **kwargs):
        return plt.figure()
    
    temporalmap_object = po.PresentationObject(dst_folder, path)
    assert isinstance(temporalmap_object.loader, po.TemporalmapLoader)
    
    monkeypatch.setattr("helpers.map_type_handling.global_ocean_plot", mockreturn)
    result = temporalmap_object.create_dict()
    cube = iris.load_cube(path)
    expected_result = {
        'title': cube.attributes['title'],
        'path': './tos_nemo_year_mean_temporalmap.gif',
        'comment': cube.attributes['comment'],
        'presentation_type': 'image',
    }
    assert result == expected_result

def test_temporalmap_map_handling_exception(tmpdir):
    dst_folder = str(tmpdir)
    path = str(tmpdir) + 'test.nc'
    cube = iris.load_cube('./tests/testdata/tos_nemo_year_mean_temporalmap.nc')
    cube.attributes['map_type'] = 'invalid type'
    iris.save(cube, path)
    temporalmap_object = po.PresentationObject(dst_folder, path)
    with pytest.raises(exceptions.InvalidMapTypeException):
        temporalmap_object.create_dict()

def test_map_map_handling_exception(tmpdir):
    dst_folder = str(tmpdir)
    path = str(tmpdir) + 'test.nc'
    cube = iris.load_cube('./tests/testdata/tos_nemo_all_mean_map.nc')
    cube.attributes['map_type'] = 'invalid type'
    iris.save(cube, path)
    map_object = po.PresentationObject(dst_folder, path)
    with pytest.raises(exceptions.InvalidMapTypeException):
        map_object.create_dict()

def test_timeseries_object(tmpdir):
    dst_folder = str(tmpdir)
    path = './tests/testdata/tos_nemo_global_mean_year_mean_timeseries.nc'
    cube = iris.load_cube(path)
    timeseries_object = po.PresentationObject(dst_folder, path)
    assert isinstance(timeseries_object.loader, po.TimeseriesLoader)
    result = timeseries_object.create_dict()
    expected_result = {
        'title': cube.attributes['title'],
        'path': './tos_nemo_global_mean_year_mean_timeseries.png',
        'comment': cube.attributes['comment'],
        'presentation_type': 'image',
    }
    assert result == expected_result

def test_scalar_object(tmpdir):
    src = ["test.yml", "test.yaml"]
    error_messages = [
        f"File not found! Ignoring {src[0]}",
        f"File not found! Ignoring {src[1]}",
    ]
    for src, msg in zip(src, error_messages):
        with pytest.raises(exceptions.PresentationException, match=msg):
            po.ScalarLoader(src).load()

    src = str(tmpdir) + "/test.yml"
    dct = {"foo": "bar", "eggs": "ham"}
    with open(src, 'w') as file:
        yaml.dump(dct, file)
    scalar_object = po.PresentationObject("", src)
    assert isinstance(scalar_object.loader, po.ScalarLoader)
    result = scalar_object.create_dict()
    assert result == {'presentation_type': 'text', **dct}


"""Processing Task that writes out the disk usage of a given folder."""
import os

from scriptengine.tasks.base.timing import timed_runner
from .scalar import Scalar

class DiskusageRteScalar(Scalar):
    """DiskusageRteScalar Processing Task"""
    def __init__(self, parameters):
        super().__init__(
            parameters={**parameters, 'value': None, 'title': None},
            required_parameters=['src']
            )

    @timed_runner
    def run(self, context):
        src = self.getarg('src', context)
        dst = self.getarg('dst', context)
        self.log_info(f"Write disk usage of {src} to {dst}")

        self.value = round(self.get_directory_size(src) * 1e-9, 1)
        self.title = "Disk Usage in GB"
        self.log_debug(f"Size of Directory: {self.value}")

        self.save(
            dst,
            title=self.title,
            comment=f"Current size of {src}.",
            value=self.value,
        )

    def get_directory_size(self, path):
        """Returns the size of `path` in Bytes."""
        self.log_debug("Getting directory size.")
        total = 0
        try:
            for entry in os.scandir(path):
                if entry.is_file():
                    total += entry.stat().st_size
                elif entry.is_dir():
                    total += self.get_directory_size(entry.path)
        except (NotADirectoryError, FileNotFoundError):
            self.log_warning(f"{path} is not a directory. Returning -1.")
            return -1e9 # gets multiplied with 1e-9 again
        except PermissionError:
            self.log_warning(f"No permission to open {path}. Returning -1.")
            return -1e9 # gets multiplied with 1e-9 again

        return total



"""Tests for scriptengine/tasks/ecearth/monitoring/scalar.py"""

from unittest.mock import patch

import pytest
import yaml

from scriptengine.tasks.ecearth.monitoring.scalar import Scalar

init = [
    {
        'title': 'Title',
        'value': 'Value',
        'comment': 'Comment',
    },
    {
        'title': 'Title',
        'value': [1, 2, 3],
    },
    {
        'title': 'Title',
        'value': '{{msg}}',
    },
]

context = [
    {
        'title': 'Title',
        'value': 'Value',
        'comment': 'Comment',
    },
    {
        'title': 'Title',
        'value': [1, 2, 3],
    },
    {
        'title': 'Title',
        'value': '{{msg}}',
        'msg': 'message',
    },
]

expected_result = [
    {
        'title': 'Title',
        'value': 'Value',
        'comment': 'Comment',
        'diagnostic_type': 'scalar',
    },
    {
        'title': 'Title',
        'value': [1, 2, 3],
        'diagnostic_type': 'scalar',
    },
    {
        'title': 'Title',
        'value': 'message',
        'diagnostic_type': 'scalar',
    },
]

@pytest.mark.parametrize("init, context, expected_result", zip(init, context, expected_result))
def test_scalar_working(tmpdir, init, context, expected_result):
    path = str(tmpdir + '/test.yml')
    init['dst'] = path
    context['dst'] = path
    scalar = Scalar(init)
    scalar.run(context)
    with open(path) as file:
        result = yaml.load(file, Loader=yaml.FullLoader)
    assert expected_result == result


def test_scalar_runtime_error(tmpdir):
    path = str(tmpdir + '/test.yml')
    init = {
        'title': 'Title',
        'comment': 'Comment',
        'dst': path,
    }
    pytest.raises(
        RuntimeError,
        Scalar,
        init,
        )

def test_scalar_extension_error(tmpdir):
    path = str(tmpdir + '/test.nc')
    init = {
        'title': 'Title',
        'comment': 'Comment',
        'value': 'Value',
        'dst': path,
    }
    context = init
    scalar = Scalar(init)
    with patch.object(scalar, 'log_warning') as mock:
        scalar.run(context)
    mock.assert_called_with((
                f"{path} does not end in valid YAML file extension. "
                f"Diagnostic will not be saved."))
    


"""Tests for scriptengine/tasks/ecearth/monitoring/temporalmap.py"""

from unittest.mock import patch

import iris

from scriptengine.tasks.ecearth.monitoring.temporalmap import Temporalmap

def test_time_map_dst_error():
    time_map = Temporalmap({})
    assert time_map.correct_file_extension("test.yml") == False



"""Base class for map processing tasks."""

import os

import iris

from scriptengine.tasks.base import Task
import helpers.file_handling as helpers

class Map(Task):
    """Map Processing Task"""

    def __init__(self, parameters, required_parameters = None):
        super().__init__(__name__, parameters, required_parameters)

    def run(self, context):
        raise NotImplementedError('Base class function Map.run() must not be called')

    def save(self, new_cube, dst):
        """save map cube in netCDF file"""
        self.log_debug(f"Saving map cube to {dst}")
        new_cube.attributes['diagnostic_type'] = 'map'
        try:
            current_cube = iris.load_cube(dst)
        except OSError: # file does not exist yet.
            iris.save(new_cube, dst)
            return

        # Iris changes metadata when saving/loading cube
        # save & reload to prevent metadata mismatch
        iris.save(new_cube, 'temp.nc')
        new_cube = iris.load_cube('temp.nc')

        current_bounds = current_cube.coord('time').bounds
        new_bounds = new_cube.coord('time').bounds
        if not current_bounds[-1][-1] > new_bounds[0][0]:
            current_cube.cell_methods = new_cube.cell_methods
            cube_list = iris.cube.CubeList([current_cube, new_cube])
            merged_cube = cube_list.merge_cube()
            simulation_avg = self.compute_simulation_avg(merged_cube)
            iris.save(simulation_avg, f"{dst}-copy.nc")
            os.remove(dst)
            os.rename(f"{dst}-copy.nc", dst)
        else:
            self.log_warning("Non-monotonic coordinate. Cube will not be saved.")
        # remove temporary save
        os.remove('temp.nc')

    def correct_file_extension(self, dst):
        """check if destination file has a valid netCDF extension"""
        if not dst.endswith(".nc"):
            self.log_warning((
                f"{dst} does not end in valid netCDF file extension. "
                f"Diagnostic will not be treated, returning now."
            ))
            return False
        return True

    def compute_simulation_avg(self, merged_cube):
        """
        Compute Time Average for the whole simulation.
        """
        self.log_debug("Computing simulation average.")
        time_weights = helpers.compute_time_weights(merged_cube, merged_cube.shape)
        simulation_avg = merged_cube.collapsed(
            'time',
            iris.analysis.MEAN,
            weights=time_weights,
        )
        return simulation_avg



"""Processing Task that calculates the annual global average of a given extensive quantity."""

import warnings

import iris

from scriptengine.tasks.base.timing import timed_runner
import helpers.file_handling as hlp
from .timeseries import Timeseries

class NemoGlobalMeanYearMeanTimeseries(Timeseries):
    """NemoGlobalMeanYearMeanTimeseries Processing Task"""
    def __init__(self, parameters):
        super().__init__(
            {**parameters, 'title': None, 'coord_value': None, 'data_value': None},
            required_parameters=['src', 'domain', 'varname']
            )

    @timed_runner
    def run(self, context):
        src = self.getarg('src', context)
        dst = self.getarg('dst', context)
        domain = self.getarg('domain', context)
        varname = self.getarg('varname', context)
        comment = (f"Global average time series of **{varname}**. "
                   f"Each data point represents the (spatial and temporal) "
                   f"average over one leg.")
        self.log_info(f"Create time series for ocean variable {varname} at {dst}.")
        self.log_debug(f"Domain: {domain}, Source file(s): {src}")

        if not self.correct_file_extension(dst):
            return

        leg_cube = hlp.load_input_cube(src, varname)

        grid = self.getarg('grid', context, default='T')
        with warnings.catch_warnings():
            # Suppress warning about insufficient metadata.
            warnings.filterwarnings(
                'ignore',
                "Collapsing a multi-dimensional coordinate.",
                UserWarning,
                )
            spatial_avg = leg_cube.collapsed(
                ['latitude', 'longitude'],
                iris.analysis.MEAN,
                weights=hlp.compute_spatial_weights(domain, leg_cube.shape, grid=grid),
                )
        # Remove auxiliary time coordinate before collapsing cube
        spatial_avg.remove_coord(spatial_avg.coord('time', dim_coords=False))
        ann_spatial_avg = spatial_avg.collapsed(
            'time',
            iris.analysis.MEAN,
            weights=hlp.compute_time_weights(spatial_avg),
        )
        # Promote time from scalar to dimension coordinate
        ann_spatial_avg = iris.util.new_axis(ann_spatial_avg, 'time')

        ann_spatial_avg = hlp.set_metadata(
            ann_spatial_avg,
            title=f'{ann_spatial_avg.long_name} (Annual Mean)',
            comment=comment,
            )

        ann_spatial_avg.cell_methods = ()
        ann_spatial_avg.add_cell_method(
            iris.coords.CellMethod('mean', coords='time', intervals='1 month')
            )
        ann_spatial_avg.add_cell_method(iris.coords.CellMethod('mean', coords='area'))

        self.save(ann_spatial_avg, dst)



"""Tests for scriptengine/tasks/ecearth/monitoring/si3_hemis_point_month_mean_temporalmap.py"""

from unittest.mock import patch

import iris

from scriptengine.tasks.ecearth.monitoring.si3_hemis_point_month_mean_temporalmap import Si3HemisPointMonthMeanTemporalmap, meta_dict

def test_si3_hemis_point_month_mean_temporalmap_once(tmpdir):
    init = {
        "src": './tests/testdata/NEMO_output_sivolu-199003.nc',
        "dst": str(tmpdir) + '/test.nc',
        "hemisphere": "north",
        "varname": "sivolu",
    }
    ice_time_map = Si3HemisPointMonthMeanTemporalmap(init)
    ice_time_map.run(init)
    cube = iris.load_cube(init['dst'])
    assert cube.attributes['map_type'] == 'polar ice sheet'
    assert cube.attributes['diagnostic_type'] == 'temporal map'

def test_si3_hemis_point_month_mean_temporalmap_twice(tmpdir):
    init_a = {
        "src": './tests/testdata/NEMO_output_sivolu-199003.nc',
        "dst": str(tmpdir) + '/test.nc',
        "hemisphere": "north",
        "varname": "sivolu",
    }
    ice_time_map = Si3HemisPointMonthMeanTemporalmap(init_a)
    ice_time_map.run(init_a)
    init_b = {
        "src": './tests/testdata/NEMO_output_sivolu-199103.nc',
        "dst": str(tmpdir) + '/test.nc',
        "hemisphere": "north",
        "varname": "sivolu",
    }
    ice_time_map = Si3HemisPointMonthMeanTemporalmap(init_b)
    ice_time_map.run(init_b)
    cube = iris.load_cube(init_b['dst'])
    assert cube.attributes['map_type'] == 'polar ice sheet'
    assert cube.attributes['diagnostic_type'] == 'temporal map'

def test_si3_hemis_point_month_mean_temporalmap_wrong_varname(tmpdir):
    init = {
        "src": './tests/testdata/NEMO_output_sivolu-199003.nc',
        "dst": str(tmpdir) + '/test.nc',
        "hemisphere": "north",
        "varname": "tos",
    }
    ice_time_map = Si3HemisPointMonthMeanTemporalmap(init)
    ice_time_map.run(init)
    with patch.object(ice_time_map, 'log_warning') as mock:
        ice_time_map.run(init)
    mock.assert_called_with((
                f"'varname' must be one of the following: {meta_dict.keys()} "
                f"Diagnostic will not be treated, returning now."))

def test_si3_hemis_point_month_mean_temporalmap_wrong_hemisphere(tmpdir):
    init = {
        "src": './tests/testdata/NEMO_output_sivolu-199003.nc',
        "dst": str(tmpdir) + '/test.nc',
        "varname": "sivolu",
        "hemisphere": "east",
    }
    ice_time_map = Si3HemisPointMonthMeanTemporalmap(init)
    ice_time_map.run(init)
    with patch.object(ice_time_map, 'log_warning') as mock:
        ice_time_map.run(init)
    mock.assert_called_with((
                f"'hemisphere' must be 'north' or 'south' but is '{init['hemisphere']}'."
                f"Diagnostic will not be treated, returning now."))

# missing: wrong dst (covered by test_time_map.py)
# missing: 'convert_to' test (requires a siconc input file) or multiple Mocks



import iris_grib._grib_cf_map
from iris_grib._grib_cf_map import G1LocalParam, CFName

additional_mappings = {
    G1LocalParam(1, 128, 98, 228): CFName('lwe_thickness_of_precipitation_amount', None, 'm'),
    G1LocalParam(1, 128, 98, 144): CFName('lwe_thickness_of_snowfall_amount', None, 'm'),
    G1LocalParam(1, 128, 98, 176): CFName('surface_net_downward_shortwave_flux', None, 'J m-2'),
    G1LocalParam(1, 128, 98, 177): CFName('surface_net_upward_longwave_flux', None, 'J m-2'),
    G1LocalParam(1, 128, 98, 178): CFName('toa_net_upward_shortwave_flux', None, 'J m-2'),
    G1LocalParam(1, 128, 98, 179): CFName('toa_outgoing_longwave_flux', None, 'J m-2'),
    G1LocalParam(1, 128, 98, 146): CFName('surface_upward_sensible_heat_flux', None, 'J m-2'),
    G1LocalParam(1, 128, 98, 147): CFName('surface_upward_latent_heat_flux', None, 'J m-2'),
}

def update_grib_mappings():
   iris_grib._grib_cf_map.GRIB1_LOCAL_TO_CF.update(additional_mappings)
   iris_grib.grib_phenom_translation._GRIB1_CF_TABLE = iris_grib.grib_phenom_translation._make_grib1_cf_table()


"""Tests for helpers/grib_cf_additions.py"""

from iris_grib.grib_phenom_translation import grib1_phenom_to_cf_info

import helpers.grib_cf_additions as additions

def test_update_grib_mappings():
    additions.update_grib_mappings()
    for key, value in additions.additional_mappings.items():
        cf_info = grib1_phenom_to_cf_info(key.t2version, key.centre, key.iParam)
        assert cf_info.standard_name == value.standard_name



"""Tests for scriptengine/tasks/ecearth/monitoring/simulatedyears_rte_scalar.py"""

from unittest.mock import patch

from scriptengine.tasks.ecearth.monitoring.simulatedyears_rte_scalar import SimulatedyearsRteScalar

def test_simulatedyears_working(tmpdir):
    test_path = str(tmpdir) + '/test.yml'
    init = {
        'dst': test_path,
        'start': "1990-01-01",
        'end': "1995-01-01"
    }
    simulated_years = SimulatedyearsRteScalar(init)
    with patch.object(simulated_years, 'save') as mock:
        simulated_years.run(init)
    mock.assert_called_with(
        init['dst'],
        title="Simulated Years",
        comment="Current number of simulated years.",
        value=5,
        )



"""Helper module for handling files."""

import os
import warnings
import yaml
import numpy as np
import iris
from iris.experimental.equalise_cubes import equalise_attributes

# Using https://stackoverflow.com/revisions/13197763/9
class ChangeDirectory:
    """Context manager for changing the current working directory"""
    def __init__(self, new_path):
        self.new_path = os.path.expanduser(new_path)
        self.saved_path = os.getcwd()

    def __enter__(self):
        os.chdir(self.new_path)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.saved_path)

def compute_spatial_weights(domain_src, array_shape, grid):
    "Compute weights for spatial averaging"
    domain_cfg = iris.load(domain_src)
    scale_factors = [
        domain_cfg.extract(f'e1{grid.lower()}')[0][0],
        domain_cfg.extract(f'e2{grid.lower()}')[0][0],
    ]
    cell_areas = scale_factors[0] * scale_factors[1]
    cell_weights = np.broadcast_to(cell_areas.data, array_shape)
    return cell_weights

def compute_time_weights(monthly_data_cube, cube_shape=None):
    """Compute weights for the different month lengths"""
    time_dim = monthly_data_cube.coord('time', dim_coords=True)
    month_weights = np.array([bound[1] - bound[0] for bound in time_dim.bounds])
    if cube_shape:
        weight_shape = np.ones(cube_shape[1:])
        month_weights = np.array([time_weight * weight_shape for time_weight in month_weights])
    return month_weights

def load_input_cube(src, varname):
    """Load input file(s) into one cube."""
    with warnings.catch_warnings():
        # Suppress psu warning
        warnings.filterwarnings(
            action='ignore',
            message="Ignoring netCDF variable",
            category=UserWarning,
            )
        month_cubes = iris.load(src, varname)
    if len(month_cubes) == 1:
        month_cube = remove_unique_attributes(month_cubes[0])
        return month_cube
    equalise_attributes(month_cubes) # 'timeStamp' and 'uuid' would cause ConcatenateError
    leg_cube = month_cubes.concatenate_cube()
    return leg_cube

def set_metadata(cube, title=None, comment=None, **kwargs):
    """Set metadata for diagnostic."""
    metadata = {
        'title': title,
        'comment': comment,
        'source': 'EC-Earth 4',
        'Conventions': 'CF-1.8',
        }
    for key, value in kwargs.items():
        metadata[f'{key}'] = value

    metadata_to_discard = [
        'description',
        'interval_operation',
        'interval_write',
        'name',
        'online_operation',
        ]
    for key, value in metadata.items():
        if value is not None:
            cube.attributes[key] = value
    for key in metadata_to_discard:
        cube.attributes.pop(key, None)
    return cube

def remove_unique_attributes(cube):
    # NEMO attributes unique to each file:
    cube.attributes.pop('uuid', None)
    cube.attributes.pop('timeStamp', None)
    return cube


"""Processing Task that creates a 2D time map of a given extensive atmosphere quantity."""

import iris
import iris_grib
import numpy as np

from scriptengine.tasks.base.timing import timed_runner

from helpers.grib_cf_additions import update_grib_mappings
import helpers.file_handling as helpers
from .temporalmap import Temporalmap

class OifsYearMeanTemporalmap(Temporalmap):
    """OifsYearMeanTemporalmap Processing Task"""

    def __init__(self, parameters):
        super().__init__(
            parameters,
            required_parameters=['src', 'dst', 'grib_code']
            )

    @timed_runner
    def run(self, context):
        src = self.getarg('src', context)
        dst = self.getarg('dst', context)
        grib_code = self.getarg('grib_code', context)
        src = [path for path in src if not path.endswith('000000')]
        self.log_info(f"Create time map for atmosphere variable {grib_code} at {dst}.")
        self.log_debug(f"Source file(s): {src}")

        if not self.correct_file_extension(dst):
            return

        update_grib_mappings()
        cf_phenomenon = iris_grib.grib_phenom_translation.grib1_phenom_to_cf_info(
            128, # table
            98, # institution: ECMWF
            grib_code
        )
        if not cf_phenomenon:
            self.log_warning(f"CF Phenomenon for {grib_code} not found. Update local table?")
            return
        self.log_debug(f"Getting variable {cf_phenomenon.standard_name}")
        leg_cube = helpers.load_input_cube(src, cf_phenomenon.standard_name)

        leg_cube.long_name = leg_cube.long_name.replace("_", " ")

        if leg_cube.units.name == 'kelvin':
            leg_cube.convert_units('degC')

        time_coord = leg_cube.coord('time')
        step = time_coord.points[1] - time_coord.points[0]
        time_coord.bounds = np.array([[point - step, point] for point in time_coord.points])

        leg_mean = leg_cube.collapsed(
            'time',
            iris.analysis.MEAN,
        )
        # Promote time from scalar to dimension coordinate
        leg_mean = iris.util.new_axis(leg_mean, 'time')

        leg_mean = self.set_cell_methods(leg_mean, step)

        leg_mean = helpers.set_metadata(
            leg_mean,
            title=f'{leg_mean.long_name.title()} (Annual Mean Map)',
            comment=f"Leg Mean of **{grib_code}**.",
            map_type="global atmosphere",
        )

        self.save(leg_mean, dst)

    def set_cell_methods(self, cube, step):
        """Set the correct cell methods."""
        cube.cell_methods = ()
        cube.add_cell_method(iris.coords.CellMethod(
            'mean',
            coords='time',
            intervals=f'{step * 3600} seconds',
            ))
        cube.add_cell_method(iris.coords.CellMethod('point', coords=['latitude', 'longitude']))
        return cube



"""Tests for scriptengine/tasks/ecearth/monitoring/si3_hemis_sum_month_mean_timeseries.py"""

from unittest.mock import patch

import iris

from scriptengine.tasks.ecearth.monitoring.si3_hemis_sum_month_mean_timeseries import Si3HemisSumMonthMeanTimeseries, meta_dict

def test_si3_hemis_sum_month_mean_timeseries_working(tmpdir):
    init = {
        "src": [
            './tests/testdata/NEMO_output_sivolu-199003.nc',
            './tests/testdata/NEMO_output_sivolu-199009.nc',
            ],
        "dst": str(tmpdir) + '/test.nc',
        "domain": './tests/testdata/domain_cfg_example.nc',
        "varname": "sivolu",
        "hemisphere": "north",
    }
    ice_time_series = Si3HemisSumMonthMeanTimeseries(init)
    ice_time_series.run(init)
    cube = iris.load_cube(str(tmpdir) + '/test.nc')
    assert cube.name() == 'sea_ice_volume'
    assert cube.attributes['title'] is not None
    assert cube.attributes['comment'] is not None
    assert cube.attributes['diagnostic_type'] == 'time series'
    assert cube.cell_methods == (
        iris.coords.CellMethod('point', coords='time'),
        iris.coords.CellMethod('sum', coords='area', intervals='northern hemisphere'),
        )

def test_si3_hemis_sum_month_mean_timeseries_wrong_varname(tmpdir):
    init = {
        "src": './tests/testdata/NEMO_output_sivolu-199003.nc',
        "dst": str(tmpdir) + '/test.nc',
        "domain": './tests/testdata/domain_cfg_example.nc',
        "varname": "tos",
        "hemisphere": "south",
    }
    ice_time_series = Si3HemisSumMonthMeanTimeseries(init)
    ice_time_series.run(init)
    with patch.object(ice_time_series, 'log_warning') as mock:
        ice_time_series.run(init)
    mock.assert_called_with((
                f"'varname' must be one of the following: {meta_dict.keys()} "
                f"Diagnostic will not be treated, returning now."))

def test_si3_hemis_sum_month_mean_timeseries_wrong_hemisphere(tmpdir):
    init = {
        "src": './tests/testdata/NEMO_output_sivolu-199003.nc',
        "dst": str(tmpdir) + '/test.nc',
        "domain": './tests/testdata/domain_cfg_example.nc',
        "varname": "sivolu",
        "hemisphere": "east",
    }
    ice_time_series = Si3HemisSumMonthMeanTimeseries(init)
    ice_time_series.run(init)
    with patch.object(ice_time_series, 'log_warning') as mock:
        ice_time_series.run(init)
    mock.assert_called_with((
                f"'hemisphere' must be 'north' or 'south' but is '{init['hemisphere']}'."
                f"Diagnostic will not be treated, returning now."))



"""Tests for helpers/map_type_handling.py"""

import iris
import matplotlib.pyplot as plt

import helpers.map_type_handling as mth

def test_global_ocean_plot():
    ocean_cube = iris.load_cube("./tests/testdata/tos_nemo_all_mean_map.nc")
    assert isinstance(mth.global_ocean_plot(ocean_cube), plt.Figure)

def test_ice_sheet_plot():
    ice_cube = iris.load_cube("./tests/testdata/sithic-north-mar.nc")
    assert isinstance(mth.polar_ice_sheet_plot(ice_cube), plt.Figure)

def test_global_atmosphere_plot():
    atmo_cube = iris.load_cube("./tests/testdata/tas-climatology.nc")
    assert isinstance(mth.global_atmosphere_plot(atmo_cube), plt.Figure)

def test_function_mapper():
    assert mth.function_mapper('global ocean') == mth.global_ocean_plot
    assert mth.function_mapper('global atmosphere') == mth.global_atmosphere_plot
    assert mth.function_mapper('polar ice sheet') == mth.polar_ice_sheet_plot
    assert mth.function_mapper("invalid") is None


""" ScriptEngine EC-Earth tasks

This module provides SE tasks for the EC-Earth ESM
"""

from .monitoring.simulatedyears_rte_scalar import SimulatedyearsRteScalar
from .monitoring.scalar import Scalar
from .monitoring.markdown import Markdown
from .monitoring.nemo_global_mean_year_mean_timeseries import NemoGlobalMeanYearMeanTimeseries
from .monitoring.diskusage_rte_scalar import DiskusageRteScalar
from .monitoring.si3_hemis_sum_month_mean_timeseries import Si3HemisSumMonthMeanTimeseries
from .monitoring.nemo_all_mean_map import NemoAllMeanMap
from .monitoring.nemo_time_mean_temporalmap import NemoYearMeanTemporalmap, NemoMonthMeanTemporalmap
from .monitoring.oifs_all_mean_map import OifsAllMeanMap
from .monitoring.oifs_year_mean_temporalmap import OifsYearMeanTemporalmap
from .monitoring.si3_hemis_point_month_mean_all_mean_map import Si3HemisPointMonthMeanAllMeanMap
from .monitoring.si3_hemis_point_month_mean_temporalmap import Si3HemisPointMonthMeanTemporalmap
from .monitoring.oifs_global_mean_year_mean_timeseries import OifsGlobalMeanYearMeanTimeseries
from .monitoring.timeseries import Timeseries
from .monitoring.redmine import Redmine
from .slurm import Sbatch

def task_loader_map():
    return {
        'sbatch': Sbatch,
        'ece.mon.scalar': Scalar,
        'ece.mon.timeseries': Timeseries,
        'ece.mon.diskusage_rte_scalar': DiskusageRteScalar,
        'ece.mon.simulatedyears_rte_scalar': SimulatedyearsRteScalar,
        'ece.mon.nemo_global_mean_year_mean_timeseries': NemoGlobalMeanYearMeanTimeseries,
        'ece.mon.nemo_all_mean_map': NemoAllMeanMap,
        'ece.mon.nemo_month_mean_temporalmap': NemoMonthMeanTemporalmap,
        'ece.mon.nemo_year_mean_temporalmap': NemoYearMeanTemporalmap,
        'ece.mon.si3_hemis_sum_month_mean_timeseries': Si3HemisSumMonthMeanTimeseries,
        'ece.mon.si3_hemis_point_month_mean_all_mean_map': Si3HemisPointMonthMeanAllMeanMap,
        'ece.mon.si3_hemis_point_month_mean_temporalmap': Si3HemisPointMonthMeanTemporalmap,
        'ece.mon.oifs_all_mean_map': OifsAllMeanMap,
        'ece.mon.oifs_year_mean_temporalmap': OifsYearMeanTemporalmap,
        'ece.mon.oifs_global_mean_year_mean_timeseries': OifsGlobalMeanYearMeanTimeseries,
        'ece.mon.presentation.markdown': Markdown,
        'ece.mon.presentation.redmine': Redmine,
        }



"""Processing Task that writes out a generalized scalar diagnostic."""

import yaml

from scriptengine.tasks.base import Task
from scriptengine.tasks.base.timing import timed_runner

class Scalar(Task):
    """Processing Task that writes out a generalized scalar diagnostic."""

    def __init__(self, parameters, required_parameters=None):
        required = [
            "title",
            "value",
            "dst",
        ]
        super().__init__(
            __name__,
            parameters,
            required_parameters=required + (required_parameters or [])
            )

    @timed_runner
    def run(self, context):
        title = self.getarg('title', context)
        dst = self.getarg('dst', context)

        self.log_info(f"Write scalar {title} to {dst}")

        value = self.getarg('value', context)
        comment = self.getarg('comment', context, default=None)

        self.save(dst, title=title, value=value, comment=comment)

    def save(self, dst, **kwargs):
        """Saves a scalar diagnostic in a YAML file."""
        self.log_debug(f"Saving scalar diagnostic to {dst}")
        filtered_dict = {k: v for k, v in kwargs.items() if v is not None}
        filtered_dict['diagnostic_type'] = 'scalar'
        if dst.endswith(".yml") or dst.endswith(".yaml"):
            with open(dst, 'w') as outfile:
                yaml.dump(filtered_dict, outfile, sort_keys=False)
        else:
            self.log_warning((
                f"{dst} does not end in valid YAML file extension. "
                f"Diagnostic will not be saved."
            ))



"""Tests for scriptengine/tasks/ecearth/monitoring/oifs_all_mean_map.py"""

from unittest.mock import patch

import iris

from scriptengine.tasks.ecearth.monitoring.oifs_all_mean_map import OifsAllMeanMap

def test_oifs_all_mean_map_working(tmpdir):
    init = {
        "src": ['./tests/testdata/OIFSGG34199001'],
        "dst": str(tmpdir) + '/test.nc',
        "grib_code": 34,
    }
    atmo_map = OifsAllMeanMap(init)
    atmo_map.run(init)
    cube = iris.load_cube(str(tmpdir) + '/test.nc')
    assert cube.name() == 'sea_surface_temperature'
    assert cube.attributes['title'] is not None
    assert cube.attributes['comment'] is not None
    assert cube.attributes['diagnostic_type'] == 'map'
    assert cube.attributes['map_type'] == 'global atmosphere'
    assert cube.coord('time').climatological
    assert len(cube.coord('time').points) == 1

def test_oifs_all_mean_map_wrong_code(tmpdir):
    init = {
        "src": ['./tests/testdata/NEMO_output_sivolu-199003.nc'],
        "dst": str(tmpdir) + '/test.nc',
        "grib_code": 500,
    }
    atmo_map = OifsAllMeanMap(init)
    with patch.object(atmo_map, 'log_warning') as mock:
        atmo_map.run(init)
    mock.assert_called_with(f"CF Phenomenon for {init['grib_code']} not found. Update local table?")



"""Tests for scriptengine/tasks/ecearth/monitoring/nemo_global_mean_year_mean_timeseries.py"""

from unittest.mock import patch

import iris

from scriptengine.tasks.ecearth.monitoring.nemo_global_mean_year_mean_timeseries import NemoGlobalMeanYearMeanTimeseries

def test_nemo_global_mean_year_mean_timeseries_working(tmpdir):
    init = {
        "src": ['./tests/testdata/NEMO_output_sivolu-199003.nc'],
        "dst": str(tmpdir) + '/test.nc',
        "domain": './tests/testdata/domain_cfg_example.nc',
        "varname": "sivolu",
        "grid": "T",
    }
    global_avg = NemoGlobalMeanYearMeanTimeseries(init)
    global_avg.run(init)
    cube = iris.load_cube(str(tmpdir) + '/test.nc')
    assert cube.name() == 'sea_ice_thickness'
    assert cube.attributes['title'] is not None
    assert cube.attributes['comment'] is not None
    assert cube.attributes['diagnostic_type'] == 'time series'
    assert cube.cell_methods == (
        iris.coords.CellMethod('mean', coords='time', intervals='1 month'),
        iris.coords.CellMethod('mean', coords='area'),
        )



"""Processing Task that writes out the number of so far simulated years."""

from dateutil.relativedelta import relativedelta

from scriptengine.tasks.base.timing import timed_runner
from .scalar import Scalar

class SimulatedyearsRteScalar(Scalar):
    """SimulatedyearsRteScalar Processing Task"""
    def __init__(self, parameters):
        super().__init__(
            parameters={**parameters, 'value': None, 'title': None},
            required_parameters=['start', 'end']
            )

    @timed_runner
    def run(self, context):
        dst = self.getarg('dst', context)
        start = self.getarg('start', context)
        end = self.getarg('end', context)
        self.log_info(f"Write simulated years to {dst}")

        self.value = relativedelta(end, start).years
        self.title = "Simulated Years"

        self.save(
            dst,
            title=self.title,
            comment="Current number of simulated years.",
            value=self.value,
        )



"""Tests for scriptengine/tasks/ecearth/monitoring/nemo_time_mean_temporalmap.py"""

import pytest
import iris

from scriptengine.tasks.ecearth.monitoring.nemo_time_mean_temporalmap import NemoTimeMeanTemporalmap, NemoMonthMeanTemporalmap, NemoYearMeanTemporalmap

def test_nemo_time_mean_temporalmap_base(tmpdir):
    init = {
        "src": [
            './tests/testdata/NEMO_output_sivolu-199003.nc',
            './tests/testdata/NEMO_output_sivolu-199009.nc',
            ],
        "dst": str(tmpdir) + '/test.nc',
        "varname": "sivolu",
    }
    ocean_time_map = NemoTimeMeanTemporalmap(init)
    with pytest.raises(NotImplementedError):
        ocean_time_map.run(init)

def test_nemo_year_mean_temporalmap_once(tmpdir):
    init = {
        "src": [
            './tests/testdata/NEMO_output_sivolu-199003.nc',
            './tests/testdata/NEMO_output_sivolu-199009.nc',
            ],
        "dst": str(tmpdir) + '/test.nc',
        "varname": "sivolu",
    }
    ocean_time_map = NemoYearMeanTemporalmap(init)
    ocean_time_map.run(init)
    cube = iris.load_cube(init['dst'])
    assert cube.attributes['map_type'] == 'global ocean'
    assert cube.attributes['diagnostic_type'] == 'temporal map'
    assert cube.data.shape[0] == 1

def test_nemo_year_mean_temporalmap_twice(tmpdir):
    init_a = {
        "src": ['./tests/testdata/NEMO_output_sivolu-199003.nc'],
        "dst": str(tmpdir) + '/test.nc',
        "varname": "sivolu",
    }
    ocean_time_map = NemoYearMeanTemporalmap(init_a)
    ocean_time_map.run(init_a)
    init_b = {
        "src": ['./tests/testdata/NEMO_output_sivolu-199009.nc'],
        "dst": str(tmpdir) + '/test.nc',
        "varname": "sivolu",
    }
    ocean_time_map = NemoYearMeanTemporalmap(init_b)
    ocean_time_map.run(init_b)
    cube = iris.load_cube(init_b['dst'])
    assert cube.attributes['map_type'] == 'global ocean'
    assert cube.attributes['diagnostic_type'] == 'temporal map'
    assert cube.data.shape[0] == 2


def test_nemo_month_mean_temporalmap_once(tmpdir):
    init = {
        "src": [
            './tests/testdata/NEMO_output_sivolu-199003.nc',
            './tests/testdata/NEMO_output_sivolu-199009.nc',
            ],
        "dst": str(tmpdir) + '/test.nc',
        "varname": "sivolu",
    }
    ocean_time_map = NemoMonthMeanTemporalmap(init)
    ocean_time_map.run(init)
    cube = iris.load_cube(init['dst'])
    assert cube.attributes['map_type'] == 'global ocean'
    assert cube.attributes['diagnostic_type'] == 'temporal map'
    assert cube.data.shape[0] == len(init['src'])

def test_nemo_month_mean_temporalmap_twice(tmpdir):
    init_a = {
        "src": ['./tests/testdata/NEMO_output_sivolu-199003.nc'],
        "dst": str(tmpdir) + '/test.nc',
        "varname": "sivolu",
    }
    ocean_time_map = NemoMonthMeanTemporalmap(init_a)
    ocean_time_map.run(init_a)
    init_b = {
        "src": ['./tests/testdata/NEMO_output_sivolu-199009.nc'],
        "dst": str(tmpdir) + '/test.nc',
        "varname": "sivolu",
    }
    ocean_time_map = NemoMonthMeanTemporalmap(init_b)
    ocean_time_map.run(init_b)
    cube = iris.load_cube(init_b['dst'])
    assert cube.attributes['map_type'] == 'global ocean'
    assert cube.attributes['diagnostic_type'] == 'temporal map'
    assert cube.data.shape[0] == len(init_a['src']) + len(init_b['src'])


"""Processing Task that creates a 2D map of a given extensive ocean quantity."""

import iris

from scriptengine.tasks.base.timing import timed_runner
import helpers.file_handling as helpers
from .map import Map

class NemoAllMeanMap(Map):
    """NemoAllMeanMap Processing Task"""

    def __init__(self, parameters):
        super().__init__(
            parameters,
            required_parameters=['src', 'dst', 'varname']
            )

    @timed_runner
    def run(self, context):
        src = self.getarg('src', context)
        dst = self.getarg('dst', context)
        varname = self.getarg('varname', context)
        self.log_info(f"Create map for ocean variable {varname} at {dst}.")
        self.log_debug(f"Source file(s): {src}")

        if not self.correct_file_extension(dst):
            return

        leg_cube = helpers.load_input_cube(src, varname)

        # Remove auxiliary time coordinate before collapsing cube
        leg_cube.remove_coord(leg_cube.coord('time', dim_coords=False))

        time_weights = helpers.compute_time_weights(leg_cube, leg_cube.shape)
        leg_average = leg_cube.collapsed(
            'time',
            iris.analysis.MEAN,
            weights=time_weights
        )

        leg_average.coord('time').climatological = True
        leg_average = self.set_cell_methods(leg_average)

        leg_average = helpers.set_metadata(
            leg_average,
            title=f'{leg_average.long_name.title()} (Annual Mean Climatology)',
            comment=f"Simulation Average of **{varname}**.",
            map_type='global ocean',
        )

        self.save(leg_average, dst)
    
    def set_cell_methods(self, cube):
        """Set the correct cell methods."""
        cube.cell_methods = ()
        cube.add_cell_method(
            iris.coords.CellMethod('mean within years', coords='time', intervals='1 month')
            )
        cube.add_cell_method(
            iris.coords.CellMethod('mean over years', coords='time')
            )
        cube.add_cell_method(
            iris.coords.CellMethod('point', coords=['latitude', 'longitude'])
            )
        return cube


"""Tests for scriptengine/tasks/ecearth/monitoring/timeseries.py"""

import datetime
from unittest.mock import patch

import iris

from scriptengine.tasks.ecearth.monitoring.timeseries import Timeseries

def test_time_series_dst_error():
    init = {
        "title": "A Test Diagnostic",
        "dst": "dst_file.yml",
        "data_value": 0,
        "coord_value": 0,
    }
    time_series = Timeseries(init)
    assert not time_series.correct_file_extension(init['dst'])

def test_monotonic_increase():
    init = {
        "title": "A Test Diagnostic",
        "dst": "dst_file.nc",
        "data_value": 0,
        "coord_value": 0,
    }
    time_series = Timeseries(init)

    old_coord = iris.coords.Coord([1])
    new_coord = iris.coords.Coord([2])
    assert time_series.test_monotonic_increase(old_coord, new_coord)
    assert not time_series.test_monotonic_increase(new_coord, old_coord)

    old_coord_with_bounds = iris.coords.Coord([2], bounds=[0.5, 1.5])
    new_coord_with_bounds = iris.coords.Coord([1], bounds=[1.5, 2.5])
    assert time_series.test_monotonic_increase(old_coord_with_bounds, new_coord_with_bounds)
    assert not time_series.test_monotonic_increase(new_coord_with_bounds, old_coord_with_bounds)

def test_time_series_first_save(tmpdir):
    init = {
        "title": "A Test Diagnostic",
        "dst": str(tmpdir) + "/dst_file.nc",
        "data_value": 0,
        "coord_value": 0,
    }
    time_series = Timeseries(init)
    time_series.run(init)

    cube = iris.load_cube(init['dst'])
    assert cube.data == [init['data_value']]
    assert cube.coord().points == [init['coord_value']]
    assert cube.attributes['title'] == init['title']
    assert cube.attributes['diagnostic_type'] == 'time series'
    assert cube.name() == init['title']
    assert cube.units.name == '1'
    assert cube.coord().name() == 'time'
    assert cube.coord().units.name == '1'

def test_time_series_append(tmpdir):
    init_a = {
        "title": "A Test Diagnostic",
        "dst": str(tmpdir) + "/dst_file.nc",
        "data_value": 0,
        "coord_value": 0,
    }
    time_series = Timeseries(init_a)
    time_series.run(init_a)

    init_b = {
        "title": "A Test Diagnostic",
        "dst": str(tmpdir) + "/dst_file.nc",
        "data_value": 0,
        "coord_value": 1,
    }

    time_series = Timeseries(init_b)
    time_series.run(init_b)

    cube = iris.load_cube(init_a['dst'])
    assert (cube.data == [init_a['data_value'], init_b['data_value']]).all()
    assert (cube.coord().points == [init_a['coord_value'], init_b['coord_value']]).all()
    assert cube.attributes['title'] == init_a['title']
    assert cube.attributes['diagnostic_type'] == 'time series'
    assert cube.name() == init_a['title']
    assert cube.coord().name() == 'time'
    assert cube.units.name == '1'
    assert cube.coord().units.name == '1'

def test_time_series_append_nonmonotonic(tmpdir):
    init_a = {
        "title": "A Test Diagnostic",
        "dst": str(tmpdir) + "/dst_file.nc",
        "data_value": 0,
        "coord_value": 1,
    }
    time_series = Timeseries(init_a)
    time_series.run(init_a)

    init_b = {
        "title": "A Test Diagnostic",
        "dst": str(tmpdir) + "/dst_file.nc",
        "data_value": 0,
        "coord_value": 0,
    }

    time_series = Timeseries(init_b)
    with patch.object(time_series, 'log_warning') as mock:
        time_series.run(init_b)
    mock.assert_called_with("Non-monotonic coordinate. Cube will not be saved.")
    


def test_time_series_date_time(tmpdir):
    seconds_value = (datetime.datetime(1990, 1, 1) - datetime.datetime(1900, 1, 1)).total_seconds()
    init_date = {
        "title": "A Date Diagnostic",
        "dst": str(tmpdir) + "/date_file.nc",
        "data_value": 0,
        "coord_value": "1990-01-01",
    }
    time_series = Timeseries(init_date)
    time_series.run(init_date)
    
    cube = iris.load_cube(init_date['dst'])
    assert cube.data == [init_date['data_value']]
    assert cube.coord().points == [seconds_value]
    assert cube.attributes['title'] == init_date['title']
    assert cube.attributes['diagnostic_type'] == 'time series'
    assert cube.name() == init_date['title']
    assert cube.units.name == '1'
    assert cube.coord().name() == 'time'
    assert cube.coord().units.name == "second since 1900-01-01 00:00:00.0000000 UTC"

    init_date_time = {
        "title": "A Datetime Diagnostic",
        "dst": str(tmpdir) + "/date_time_file.nc",
        "data_value": 0,
        "coord_value": "1990-01-01 00:00:00",
    }
    time_series = Timeseries(init_date_time)
    time_series.run(init_date_time)
    
    cube = iris.load_cube(init_date_time['dst'])
    assert cube.data == [init_date_time['data_value']]
    assert cube.coord().points == [seconds_value]
    assert cube.attributes['title'] == init_date_time['title']
    assert cube.attributes['diagnostic_type'] == 'time series'
    assert cube.name() == init_date_time['title']
    assert cube.units.name == '1'
    assert cube.coord().name() == 'time'
    assert cube.coord().units.name == "second since 1900-01-01 00:00:00.0000000 UTC"



"""Processing Task that writes out a generalized time series diagnostic."""

import os
import datetime

import iris
import numpy as np

from scriptengine.tasks.base import Task
from scriptengine.tasks.base.timing import timed_runner
import helpers.file_handling as helpers

class Timeseries(Task):
    """Processing Task that writes out a generalized time series diagnostic."""

    def __init__(self, parameters, required_parameters = None):
        required = [
            "title",
            "coord_value",
            "data_value",
            "dst",
        ]
        super().__init__(
            __name__,
            parameters,
            required_parameters=required + (required_parameters or [])
            )

    @timed_runner
    def run(self, context):
        # load input parameters
        title = self.getarg('title', context)
        dst = self.getarg('dst', context)
        self.log_info(f"Time series {title} at {dst}.")

        # Convert coordinate value to number and get unit
        coord_value = self.getarg("coord_value", context, default=None)
        if type(coord_value) in (datetime.datetime, datetime.date):
            try:
                coord_value = (coord_value - datetime.datetime(1900, 1, 1)).total_seconds()
            except TypeError:
                coord_value = (coord_value - datetime.date(1900, 1, 1)).total_seconds()
            coord_unit = "second since 1900-01-01 00:00:00"
        else:
            coord_unit = self.getarg("coord_unit", context, default="1")
        
        data_value = self.getarg("data_value", context)
        data_unit = self.getarg("data_unit", context, default="1")

        coord_name = self.getarg('coord_name', context, default='time')
        data_name = self.getarg('data_name', context, default=title)
        comment = self.getarg('comment', context, default=".")
        
        self.log_debug(f"Value: {data_value} at time: {coord_value}, title: {title}")

        if not self.correct_file_extension(dst):
            return

        # create coord
        coord = iris.coords.DimCoord(
            points=np.array([coord_value]),
            long_name=coord_name,
            var_name=coord_name.replace(" ", "_"),
            units=coord_unit,
        )

        # create cube
        data_cube = iris.cube.Cube(
            data=np.array([data_value]),
            long_name=data_name,
            var_name=data_name.replace(" ", "_"),
            units=data_unit,
            dim_coords_and_dims=[(coord, 0)],
        )

        # set metadata
        data_cube = helpers.set_metadata(
            data_cube,
            title=title,
            comment=comment,
        )
        self.save(data_cube, dst)

    def save(self, new_cube, dst):
        """save time series cube in netCDF file"""
        self.log_debug(f"Saving time series cube to {dst}")

        new_cube.attributes['diagnostic_type'] = 'time series'
        try:
            current_cube = iris.load_cube(dst)
        except OSError: # file does not exist yet.
            iris.save(new_cube, dst)
            return

        # Iris changes metadata when saving/loading cube
        # save & reload to prevent metadata mismatch
        iris.save(new_cube, 'temp.nc')
        new_cube = iris.load_cube('temp.nc')

        if self.test_monotonic_increase(current_cube.coords()[0], new_cube.coords()[0]):
            cube_list = iris.cube.CubeList([current_cube, new_cube])
            merged_cube = cube_list.concatenate_cube()
            iris.save(merged_cube, f"{dst}-copy.nc")
            os.remove(dst)
            os.rename(f"{dst}-copy.nc", dst)
        else:
            self.log_warning("Non-monotonic coordinate. Cube will not be saved.")

        # remove temporary save
        os.remove('temp.nc')

    def test_monotonic_increase(self, old_coord, new_coord):
        """Test if coordinate is monotonically increasing."""
        current_bounds = old_coord.bounds
        new_bounds = new_coord.bounds
        if current_bounds is not None and new_bounds is not None:
            if current_bounds[-1][-1] > new_bounds[0][0]:
                self.log_warning("Inserting would lead to non-monotonic time axis.")
                return False
            return True

        if old_coord.points[-1] > new_coord.points[0]:
            self.log_warning("Inserting would lead to non-monotonic time axis.")
            return False
        return True

    def correct_file_extension(self, dst):
        """check if destination file has a valid netCDF extension"""
        if not dst.endswith(".nc"):
            self.log_warning((
                f"{dst} does not end in valid netCDF file extension. "
                f"Diagnostic will not be treated, returning now."
            ))
            return False
        return True



"""SLURM Sbatch task for ScriptEngine."""

import os
import subprocess

from scriptengine.tasks.base import Task
from scriptengine.jinja import render as j2render
from scriptengine.exceptions import ScriptEngineStopException


class Sbatch(Task):
    """Sbatch task, submits SE scripts to the SLURM batch system
    """

    def __init__(self, parameters):
        super().__init__(__name__, parameters, required_parameters=["scripts"])

    def run(self, context):

        # Default is to _not_ submit a job if we are already within a batch job
        submit_from_batch_job = False

        # Get all arguments from the task description
        # Almost all arguments are just passed to the 'sbatch' command, except
        # for the 'scripts' and 'submit_from_batch_job', which are used here
        opt_args = []
        for opt, arg in self.__dict__.items():
            if opt == "scripts":
                if isinstance(arg, list):
                    scripts = [j2render(script, context) for script in arg]
                else:
                    scripts = [j2render(arg, context)]
                self.log_info(f"Submit job for {scripts}")
            elif opt == "submit_from_batch_job":
                submit_from_batch_job = arg
            elif not opt.startswith("_"):
                opt_args.append(f"--{opt}")
                if arg:
                    opt_args.append(j2render(arg, context))

        # Check if we are within a SLURM batch job and if we want to submit a
        # job from within a job
        if os.environ.get("SLURM_JOB_NAME"):
            if not submit_from_batch_job:
                self.log_debug("Sbatch task from within a SLURM batch job "
                               "and 'submit_from_batch_job' is not set or "
                               "false: not submitting new job")
                return

        # Build the sbatch command line
        sbatch_cmd = ["sbatch"]
        sbatch_cmd.extend(map(str, opt_args))
        sbatch_cmd.append("se")
        sbatch_cmd.extend(map(str, scripts))

        self.log_debug(f"Sbatch command line: {sbatch_cmd}")

        # Run sbatch command, with handling of errors
        try:
            subprocess.run(sbatch_cmd, check=True)
        except subprocess.CalledProcessError:
            self.log_error("Sbatch command returns an error")
            raise
        else:
            raise ScriptEngineStopException("Sbatch task requests stop "
                                            "after submitting batch job")



"""Processing Task that creates a temporal map of a given 2D extensive ocean quantity."""

import iris

from scriptengine.tasks.base.timing import timed_runner
import helpers.file_handling as helpers
from .temporalmap import Temporalmap

class NemoTimeMeanTemporalmap(Temporalmap):
    """NemoTimeMeanTemporalmap Processing Task"""

    def __init__(self, parameters, required_parameters = None):
        super().__init__(
            parameters,
            required_parameters=['src', 'dst', 'varname'] + (required_parameters or [])
            )

    @timed_runner
    def run(self, context):
        src = self.getarg('src', context)
        dst = self.getarg('dst', context)
        varname = self.getarg('varname', context)
        self.log_info(f"Create temporal map for ocean variable {varname} at {dst}.")
        self.log_debug(f"Source file(s): {src}")

        if not self.correct_file_extension(dst):
            return

        leg_cube = helpers.load_input_cube(src, varname)

        # Remove auxiliary time coordinate before collapsing cube
        leg_cube.remove_coord(leg_cube.coord('time', dim_coords=False))

        processed_cube = self.time_operation(varname, leg_cube)
        self.save(processed_cube, dst)
    
    def time_operation(self, varname, leg_cube):
        raise NotImplementedError('Base class function NemoTimeMeanTemporalmap.time_operation() must not be called')

class NemoYearMeanTemporalmap(NemoTimeMeanTemporalmap):
    """NemoYearMeanTemporalmap Processing Task"""

    def __init__(self, parameters):
        super().__init__(parameters)

    def time_operation(self, varname, leg_cube):
        self.log_debug("Creating an annual mean.")
        month_weights = helpers.compute_time_weights(leg_cube, leg_cube.shape)
        leg_average = leg_cube.collapsed(
            'time',
            iris.analysis.MEAN,
            weights=month_weights
        )
        # Promote time from scalar to dimension coordinate
        leg_average = iris.util.new_axis(leg_average, 'time')
        leg_average = helpers.set_metadata(
            leg_average,
            title=f'{leg_average.long_name.title()} (Annual Mean Map)',
            comment=f"Leg Mean of **{varname}**.",
            map_type="global ocean",
        )
        leg_average.cell_methods = ()
        leg_average.add_cell_method(iris.coords.CellMethod('mean', coords='time', intervals='1 month'))
        leg_average.add_cell_method(iris.coords.CellMethod('point', coords=['latitude', 'longitude']))
        return leg_average

class NemoMonthMeanTemporalmap(NemoTimeMeanTemporalmap):
    """NemoMonthMeanTemporalmap Processing Task"""

    def __init__(self, parameters):
        super().__init__(parameters)

    def time_operation(self, varname, leg_cube):
        self.log_debug("Creating monthly means.")
        leg_cube = helpers.set_metadata(
            leg_cube,
            title=f'{leg_cube.long_name.title()} (Monthly Mean Map)',
            comment=f"Monthly Mean of **{varname}**.",
            map_type='global ocean',
        )
        leg_cube.add_cell_method(iris.coords.CellMethod('point', coords=['latitude', 'longitude']))
        return leg_cube



"""setup.py for package scriptengine-tasks-ecearth."""
import os
import codecs
import setuptools

def read(rel_path):
    """
    Helper function to read file in relative path.
    """
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()

def get_version(rel_path):
    """
    Helper function to get package version.
    """
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]

setuptools.setup(
    name="scriptengine-tasks-ecearth",
    version=get_version("scriptengine/tasks/ecearth/version.py"),
    author="Uwe Fladrich, Valentina Schueller",
    author_email="uwe.fladrich@protonmail.com, valentina.schueller@gmail.com",
    description="ScriptEngine tasks for use with the EC-Earth climate model",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/uwefladrich/scriptengine-tasks-ecearth",
    packages=[
        "helpers",
        "scriptengine.tasks.ecearth",
        "scriptengine.tasks.ecearth.monitoring",
        "tests",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "scriptengine>=0.6",
        "pyYAML>=5.1",
        "numpy>=1.16.1",
        "imageio>=2.0",
        "scitools-iris>=2.4",
        "cartopy>=0.18",
        "iris-grib>=0.15",
        "python-redmine",
    ],
)



"""Presentation Task that saves Data and Plots to a Markdown File."""

import os

import jinja2
import yaml
import iris

from scriptengine.tasks.base import Task
from scriptengine.tasks.base.timing import timed_runner
from scriptengine.jinja import filters as j2filters
from helpers.file_handling import ChangeDirectory
from helpers.presentation_objects import PresentationObject
from helpers.exceptions import PresentationException

class Markdown(Task):
    """Markdown Presentation Task"""
    def __init__(self, parameters):
        required = [
            "src",
            "dst",
            "template",
        ]
        super().__init__(__name__, parameters, required_parameters=required)

    @timed_runner
    def run(self, context):
        sources = self.getarg('src', context)
        dst_folder = self.getarg('dst', context)
        template_path = self.getarg('template', context)
        self.log_info(f"Create Markdown report at {dst_folder}.")
        self.log_debug(f"Template: {template_path}, Source File(s): {sources}")

        presentation_list = self.get_presentation_list(sources, dst_folder)
        md_template = self.get_template(context, template_path)

        with ChangeDirectory(dst_folder):
            with open("./summary.md", 'w') as md_out:
                md_out.write(md_template.render(
                    presentation_list=presentation_list,
                ))

    def get_presentation_list(self, sources, dst_folder):
        """create a list of presentation objects"""
        self.log_debug("Getting list of presentation objects.")
        presentation_list = []
        for src in sources:
            try:
                try:
                    pres_object = PresentationObject(dst_folder, **src)
                except TypeError:
                    pres_object = PresentationObject(dst_folder, src)
                self.log_debug(f"Loading {pres_object.loader.diag_type} diagnostic from {pres_object.loader.path}.")
                presentation_list.append(pres_object.create_dict())
            except PresentationException as msg:
                self.log_warning(f"Can not present diagnostic: {msg}")
        return presentation_list

    def get_template(self, context, template_path):
        """get Jinja2 template file"""
        search_path = ['.', 'templates']
        if "_se_cmd_cwd" in context:
            search_path.extend([context["_se_cmd_cwd"],
                                os.path.join(context["_se_cmd_cwd"], "templates")])
        self.log_debug(f"Search path for template: {search_path}")

        loader = jinja2.FileSystemLoader(search_path)
        environment = jinja2.Environment(loader=loader)
        for name, function in j2filters().items():
            environment.filters[name] = function
        return environment.get_template(template_path)



"""Tests for scriptengine/tasks/monitoring/redmine.py"""

from unittest.mock import patch, Mock

import pytest
import iris
import yaml
import redminelib
import redminelib.exceptions
import jinja2.exceptions

from scriptengine.tasks.ecearth.monitoring.redmine import Redmine

def test_redmine_presentation_list(tmpdir):
    init = {
        "src": str(tmpdir),
        "local_dst": str(tmpdir),
        "template": str(tmpdir) + "/redmine_template.txt.j2",
        "subject": "Test Issue",
        "api_key": "Invalid Key"
    }
    redmine_task = Redmine(init)
    test_sources = [
        'path.txt',
        {'path': 'path.txt'},
        'path.yml'
    ]
    with patch.object(redmine_task, 'log_warning') as mock:
        result = redmine_task.get_presentation_list(test_sources, str(tmpdir))
    mock.assert_called_with("Can not present diagnostic: File not found! Ignoring path.yml")
    assert result == []

def test_redmine_connection_error(tmpdir):
    init = {
        "src": str(tmpdir),
        "local_dst": str(tmpdir),
        "template": str(tmpdir) + "/redmine_template.txt.j2",
        "subject": "Test Issue",
        "api_key": "Invalid Key"
    }
    redmine_task = Redmine(init)
    redmine = redminelib.Redmine('https://www.invalid.url', key=init['api_key'])
    with patch.object(redmine_task, 'log_warning') as mock:
       redmine_task.get_issue(redmine, init['subject'])
    mock.assert_called_with("Could not log in to Redmine server (ConnectionError)")

def test_redmine_auth_error(tmpdir):
    init = {
        "src": str(tmpdir),
        "local_dst": str(tmpdir),
        "template": str(tmpdir) + "/redmine_template.txt.j2",
        "subject": "Test Issue",
        "api_key": "Invalid Key"
    }
    redmine_task = Redmine(init)
    redmine = redminelib.Redmine('https://dev.ec-earth.org/', key=init['api_key'])
    with patch.object(redmine_task, 'log_warning') as mock:
       redmine_task.get_issue(redmine, init['subject'])
    mock.assert_called_with("Could not log in to Redmine server (AuthError)")

def test_redmine_get_template(tmpdir):
    init = {
        "src": str(tmpdir),
        "local_dst": str(tmpdir),
        "template": str(tmpdir) + "/template.txt.j2",
        "subject": "Test Issue",
        "api_key": "Invalid Key"
    }
    redmine_output = Redmine(init)
    with pytest.raises(jinja2.exceptions.TemplateNotFound):
        redmine_output.get_template(init, init['template'])

class MockTemplateOrIssue:
    def __init__(self):
        self.globals = {}

    def render(self, **kwargs):
        return str(kwargs)
    
    def save(self):
        return None

def test_redmine_run(tmpdir):
    init = {
        "src": str(tmpdir),
        "local_dst": str(tmpdir),
        "template": str(tmpdir) + "/template.txt.j2",
        "subject": "Test Issue",
        "api_key": "Invalid Key"
    }
    redmine_output = Redmine(init)
    redmine_output.get_presentation_list = Mock(return_value=[])
    mock_object = MockTemplateOrIssue()
    redmine_output.get_template = Mock(return_value=mock_object)
    redmine_output.get_issue = Mock(return_value=mock_object)
    with patch.object(redmine_output, 'log_debug') as mock:
       redmine_output.run(init)
    mock.assert_called_with("Saving issue.")



"""Processing Task that creates a 2D map of a given extensive atmosphere quantity."""

import iris
import iris_grib
import numpy as np

from scriptengine.tasks.base.timing import timed_runner
from helpers.grib_cf_additions import update_grib_mappings
import helpers.file_handling as helpers
from .map import Map

class OifsAllMeanMap(Map):
    """OifsAllMeanMap Processing Task"""

    def __init__(self, parameters):
        super().__init__(
            parameters,
            required_parameters=['src', 'dst', 'grib_code']
            )

    @timed_runner
    def run(self, context):
        src = self.getarg('src', context)
        dst = self.getarg('dst', context)
        grib_code = self.getarg('grib_code', context)
        self.log_info(f"Create map for atmosphere variable {grib_code} at {dst}.")
        src = [path for path in src if not path.endswith('000000')]
        self.log_debug(f"Source file(s): {src}")

        if not self.correct_file_extension(dst):
            return

        update_grib_mappings()
        cf_phenomenon = iris_grib.grib_phenom_translation.grib1_phenom_to_cf_info(
            128, # table
            98, # institution: ECMWF
            grib_code
        )
        if not cf_phenomenon:
            self.log_warning(f"CF Phenomenon for {grib_code} not found. Update local table?")
            return
        self.log_debug(f"Getting variable {cf_phenomenon.standard_name}")
        leg_cube = helpers.load_input_cube(src, cf_phenomenon.standard_name)

        time_coord = leg_cube.coord('time')
        step = time_coord.points[1] - time_coord.points[0]
        time_coord.bounds = np.array([[point - step, point] for point in time_coord.points])

        leg_mean = leg_cube.collapsed(
            'time',
            iris.analysis.MEAN,
        )
        leg_mean.coord('time').climatological = True
        leg_mean.cell_methods = ()
        leg_mean.add_cell_method(iris.coords.CellMethod('mean within years', coords='time', intervals=f'{step * 3600} seconds'))
        leg_mean.add_cell_method(iris.coords.CellMethod('mean over years', coords='time'))
        leg_mean.add_cell_method(iris.coords.CellMethod('point', coords=['latitude', 'longitude']))

        leg_mean.long_name = leg_mean.long_name.replace("_", " ")

        leg_mean = helpers.set_metadata(
            leg_mean,
            title=f'{leg_mean.long_name.title()} (Annual Mean Climatology)',
            comment=f"Simulation Average of **{grib_code}**.",
            map_type='global atmosphere',
        )
        if leg_mean.units.name == 'kelvin':
            leg_mean.convert_units('degC')

        self.save(leg_mean, dst)



"""Tests for scriptengine/tasks/ecearth/monitoring/diskusage_rte_scalar.py"""

from unittest.mock import patch

import yaml

from scriptengine.tasks.ecearth.monitoring.diskusage_rte_scalar import DiskusageRteScalar

def test_diskusage_working(tmpdir):
    """test full execution of DiskusageRteScalar"""
    test_path = str(tmpdir)
    subfile = tmpdir.mkdir("sub").join("file.txt")
    subfile.write("content")
    path = test_path + 'test.yml'
    init = {
        'src': test_path,
        'dst': path,
    }
    context = init
    disk_usage = DiskusageRteScalar(init)
    disk_usage.run(context)
    with open(path) as file:
        result = yaml.load(file, Loader=yaml.FullLoader)
    assert result['diagnostic_type'] == 'scalar'
    assert result['value'] == 0.0 # DiskUsage rounds to GB


def test_not_a_directory(tmpdir):
    """test with erroneous input: file instead of directory"""
    path = tmpdir.mkdir("sub").join("hello.yml")
    path.write("content")
    init = {
        'src': str(path),
        'dst': str(path),
    }
    context = init
    disk_usage = DiskusageRteScalar(init)
    disk_usage.run(context)
    expected_result = {
        'title': "Disk Usage in GB",
        'diagnostic_type': 'scalar',
        'value': -1,
    }
    with patch.object(disk_usage, 'log_warning') as mock:
        disk_usage.run(context)
    mock.assert_called_with(f"{path} is not a directory. Returning -1.")
    with open(path) as file:
        result = yaml.load(file, Loader=yaml.FullLoader)
    assert expected_result['title'] == result['title']
    assert expected_result['diagnostic_type'] == result['diagnostic_type']
    assert expected_result['value'] == result['value']

def test_non_existent_directory(tmpdir):
    """test with erroneous input: directory does not exist"""
    init = {
        'src': str(tmpdir) + "/non-existent",
        'dst': str(tmpdir) + ".yml",
    }
    context = init
    disk_usage = DiskusageRteScalar(init)
    disk_usage.run(context)
    expected_result = {
        'title': "Disk Usage in GB",
        'diagnostic_type': 'scalar',
        'value': -1,
    }
    with patch.object(disk_usage, 'log_warning') as mock:
        disk_usage.run(context)
    mock.assert_called_with(f"{init['src']} is not a directory. Returning -1.")
    with open(init['dst']) as file:
        result = yaml.load(file, Loader=yaml.FullLoader)
    assert expected_result['title'] == result['title']
    assert expected_result['diagnostic_type'] == result['diagnostic_type']
    assert expected_result['value'] == result['value']

def test_permission_error(tmpdir):
    """test with erroneous input: permission error"""
    path = str(tmpdir) + '/test.yml'
    forbidden = '/root'
    init = {
        'src': forbidden,
        'dst': str(path),
    }
    context = init
    disk_usage = DiskusageRteScalar(init)
    disk_usage.run(context)
    expected_result = {
        'title': "Disk Usage in GB",
        'diagnostic_type': 'scalar',
        'value': -1,
    }
    with patch.object(disk_usage, 'log_warning') as mock:
        disk_usage.run(context)
    mock.assert_called_with(f"No permission to open {forbidden}. Returning -1.")
    with open(path) as file:
        result = yaml.load(file, Loader=yaml.FullLoader)
    assert expected_result['title'] == result['title']
    assert expected_result['diagnostic_type'] == result['diagnostic_type']
    assert expected_result['value'] == result['value']



"""Processing Task that creates a 2D map of a given extensive ocean quantity."""

import os
import warnings

import iris
import iris_grib
import numpy as np

from scriptengine.tasks.base.timing import timed_runner
from helpers.grib_cf_additions import update_grib_mappings
import helpers.file_handling as helpers
from .timeseries import Timeseries

class OifsGlobalMeanYearMeanTimeseries(Timeseries):
    """OifsGlobalMeanYearMeanTimeseries Processing Task"""
    def __init__(self, parameters):
        super().__init__(
            {**parameters, 'title': None, 'coord_value': None, 'data_value': None},
            required_parameters=['src', 'grib_code']
            )

    @timed_runner
    def run(self, context):
        src = self.getarg('src', context)
        dst = self.getarg('dst', context)
        grib_code = self.getarg('grib_code', context)
        src = [path for path in src if not path.endswith('000000')]
        self.log_info(f"Create time series for atmosphere variable {grib_code} at {dst}.")
        self.log_debug(f"Source file(s): {src}")

        if not self.correct_file_extension(dst):
            return

        update_grib_mappings()
        cf_phenomenon = iris_grib.grib_phenom_translation.grib1_phenom_to_cf_info(
            128, # table
            98, # institution: ECMWF
            grib_code
        )
        if not cf_phenomenon:
            self.log_warning(f"CF Phenomenon for {grib_code} not found. Update local table?")
            return
        self.log_debug(f"Getting variable {cf_phenomenon.standard_name}")
        leg_cube = helpers.load_input_cube(src, cf_phenomenon.standard_name)

        time_coord = leg_cube.coord('time')
        step = time_coord.points[1] - time_coord.points[0]
        time_coord.bounds = np.array([[point - step, point] for point in time_coord.points])

        self.log_debug("Averaging over the leg.")
        leg_mean = leg_cube.collapsed(
            'time',
            iris.analysis.MEAN,
        )

        area_weights = self.get_area_weights(leg_mean)
        self.log_debug("Averaging over space.")
        with warnings.catch_warnings():
            # Suppress warning about insufficient metadata.
            warnings.filterwarnings(
                'ignore',
                "Collapsing a non-contiguous coordinate.",
                UserWarning,
                )
            spatial_mean = leg_mean.collapsed(
                ['latitude', 'longitude'],
                iris.analysis.MEAN,
                weights=area_weights,
            )
        
        spatial_mean.cell_methods = ()
        spatial_mean.add_cell_method(iris.coords.CellMethod('mean', coords='time', intervals=f'{step * 3600} seconds'))
        spatial_mean.add_cell_method(iris.coords.CellMethod('mean', coords='area'))

        # Promote time from scalar to dimension coordinate
        spatial_mean = iris.util.new_axis(spatial_mean, 'time')

        spatial_mean.long_name = spatial_mean.long_name.replace("_", " ")

        comment = (f"Global average time series of **{grib_code}**. "
                   f"Each data point represents the (spatial and temporal) "
                   f"average over one leg.")
        spatial_mean = helpers.set_metadata(
            spatial_mean,
            title=f'{spatial_mean.long_name} (Annual Mean)',
            comment=comment,
        )

        if spatial_mean.units.name == 'kelvin':
            spatial_mean.convert_units('degC')
        self.save(spatial_mean, dst)

    def get_area_weights(self, cube):
        """compute area weights for the reduced gaussian grid"""
        self.log_debug("Getting area weights.")
        nh_latitudes = np.ma.masked_less(cube.coord('latitude').points, 0)
        unique_lats, gridpoints_per_lat = np.unique(nh_latitudes, return_counts=True)
        unique_lats, gridpoints_per_lat = unique_lats[0:-1], gridpoints_per_lat[0:-1]
        areas = []
        last_angle = 0
        earth_radius = 6371
        for latitude, amount in zip(unique_lats, gridpoints_per_lat):
            delta = latitude - last_angle
            current_angle = last_angle + 2 * delta
            sin_diff = np.sin(np.deg2rad(current_angle)) - np.sin(np.deg2rad(last_angle))
            ring_area = 2 * np.pi * earth_radius**2 * sin_diff
            grid_area = ring_area / amount
            areas.extend([grid_area] * amount)
            last_angle = current_angle
        areas = np.append(areas[::-1], areas)
        area_weights = np.broadcast_to(areas, cube.data.shape)
        return area_weights
